#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.07.30 00:00:00                  #
# ================================================== #

import base64
import os
import re
from typing import Optional, Union, Dict, Any, List

from pygpt_net.core.types import (
    MODE_VISION,
)
from pygpt_net.core.bridge.context import BridgeContext
from pygpt_net.item.attachment import AttachmentItem
from pygpt_net.item.ctx import CtxItem
from pygpt_net.item.model import ModelItem


class Vision:
    def __init__(self, window=None):
        """
        Vision wrapper

        :param window: Window instance
        """
        self.window = window
        self.attachments = {}
        self.urls = []
        self.input_tokens = 0

    def send(
            self,
            context: BridgeContext,
            extra: Optional[Dict[str, Any]] = None
    ):
        """
        Call OpenAI API for chat with vision

        :param context: Bridge context
        :param extra: Extra arguments
        :return: response or stream chunks
        """
        prompt = context.prompt
        stream = context.stream
        max_tokens = int(context.max_tokens or 0)
        system_prompt = context.system_prompt
        attachments = context.attachments
        model = context.model
        model_id = model.id
        client = self.window.core.api.openai.get_client()

        # extra API kwargs
        response_kwargs = {}
        if max_tokens > 0:
            response_kwargs['max_tokens'] = max_tokens

        # build chat messages
        messages = self.build(
            prompt=prompt,
            system_prompt=system_prompt,
            model=model,
            history=context.history,
            attachments=attachments,
        )
        response = client.chat.completions.create(
            messages=messages,
            model=model_id,
            stream=stream,
            **response_kwargs
        )

        return response

    def build(
            self,
            prompt: str,
            system_prompt: str,
            model: ModelItem,
            history: Optional[List[CtxItem]] = None,
            attachments: Optional[Dict[str, AttachmentItem]] = None,
    ) -> List[dict]:
        """
        Build chat messages list

        :param prompt: user prompt
        :param system_prompt: system prompt
        :param model: model item
        :param history: history
        :param attachments: attachments
        :return: messages list
        """
        messages = []

        # tokens config
        mode = MODE_VISION
        used_tokens = self.window.core.tokens.from_user(
            prompt,
            system_prompt,
        )  # threshold and extra included
        max_ctx_tokens = self.window.core.config.get('max_total_tokens')

        # fit to max model tokens
        if max_ctx_tokens > model.ctx and model.ctx > 0:
            max_ctx_tokens = model.ctx

        # input tokens: reset
        self.reset_tokens()

        # append initial (system) message
        if system_prompt is not None and system_prompt != "":
            messages.append({
                "role": "system",
                "content": system_prompt,
            })
        else:
            if system_prompt is not None and system_prompt != "":
                messages.append({
                    "role": "system",
                    "content": system_prompt,
                })

        # append messages from context (memory)
        if self.window.core.config.get('use_context'):
            items = self.window.core.ctx.get_history(
                history,
                model.id,
                mode,
                used_tokens,
                max_ctx_tokens,
            )
            for item in items:
                # input
                if item.final_input is not None and item.final_input != "":
                    messages.append({
                        "role": "user",
                        "content": item.final_input,
                    })

                # output
                if item.final_output is not None and item.final_output != "":
                    messages.append({
                        "role": "assistant",
                        "content": item.final_output,
                    })

        # append current prompt
        content = self.build_content(prompt, attachments)
        messages.append({
            "role": "user",
            "content": content,
        })

        # input tokens: update
        self.input_tokens += self.window.core.tokens.from_messages(
            messages,
            model.id,
        )
        return messages

    def build_content(
            self,
            content: Union[str, list],
            attachments: Optional[Dict[str, AttachmentItem]] = None,
            responses_api: Optional[bool] = False,
    ) -> List[dict]:
        """
        Build vision content

        :param content: content (str or list)
        :param attachments: attachments (dict, optional)
        :param responses_api: if True, use responses API format
        :return: List of contents
        """
        type_text = "text"
        type_image = "image_url"
        if responses_api:
            type_text = "input_text"
            type_image = "input_image"

        if not isinstance(content, list):
            content = [
                {
                    "type": type_text,
                    "text": str(content)
                }
            ]

        prompt = content[0]['text']

        self.attachments = {}  # reset attachments, only current prompt
        self.urls = []

        # extract URLs from prompt
        urls = self.extract_urls(prompt)
        if len(urls) > 0:
            for url in urls:
                if not responses_api:
                    content.append(
                        {
                            "type": type_image,
                            "image_url": {
                                "url": url,
                            }
                        }
                    )
                else:
                    content.append(
                        {
                            "type": type_image,
                            "image_url": url,
                        }
                    )
                self.urls.append(url)

        # local images (attachments)
        if attachments is not None and len(attachments) > 0:
            for id in attachments:
                attachment = attachments[id]
                if os.path.exists(attachment.path):
                    # check if it's an image
                    if self.is_image(attachment.path):
                        base64_image = self.encode_image(attachment.path)
                        if not responses_api:
                            content.append(
                                {
                                    "type": type_image,
                                    "image_url": {
                                        "url": f"data:image/jpeg;base64,{base64_image}",
                                    }
                                }
                            )
                        else:
                            content.append(
                                {
                                    "type": type_image,
                                    "image_url": f"data:image/jpeg;base64,{base64_image}",
                                }
                            )
                        self.attachments[id] = attachment.path
                        attachment.consumed = True

        return content

    def build_agent_input(
            self,
            prompt: str,
            attachments: Optional[Dict[str, AttachmentItem]] = None,
    ) -> List[dict]:
        """
        Build agent input content

        :param prompt: user prompt
        :param attachments: attachments (dict, optional)
        :return: List of contents
        """
        items = []
        content = []
        self.attachments = {}  # reset attachments, only current prompt
        self.urls = []

        # extract URLs from prompt
        urls = self.extract_urls(prompt)
        if len(urls) > 0:
            for url in urls:
                content.append(
                    {
                        "type": "input_image",
                        "image_url": url,
                    }
                )
                self.urls.append(url)

        # local images (attachments)
        if attachments is not None and len(attachments) > 0:
            for id in attachments:
                attachment = attachments[id]
                if os.path.exists(attachment.path):
                    # check if it's an image
                    if self.is_image(attachment.path):
                        base64_image = self.encode_image(attachment.path)
                        content.append(
                            {
                                "type": "input_image",
                                "detail": "auto",
                                "image_url": f"data:image/jpeg;base64,{base64_image}",
                            }
                        )
                        self.attachments[id] = attachment.path
                        attachment.consumed = True

        if content:
            items.append({
                "role": "user",
                "content": content
            })
        items.append({
            "role": "user",
            "content": str(prompt)
        })

        return items

    def get_attachment(
            self,
            attachments: Optional[Dict[str, AttachmentItem]] = None,
    ) -> str:
        """
        Get attachment as base64 encoded image

        :param attachments: attachments dict
        :return: base64 encoded image or empty string if no image found
        """
        if attachments is not None and len(attachments) > 0:
            for id in attachments:
                attachment = attachments[id]
                if os.path.exists(attachment.path):
                    # check if it's an image
                    if self.is_image(attachment.path):
                        base64_image = self.encode_image(attachment.path)
                        attachment.consumed = True
                        return base64_image

    def extract_urls(self, text: str) -> List[str]:
        """
        Extract img urls from text

        :param text: text
        :return: list of img urls
        """
        urls = re.findall(r'(https?://\S+)', text)
        img_urls = []
        for url in urls:
            if self.is_image(url):
                img_urls.append(url)

        # if single url, return it
        if "".join(urls).strip() == text.strip():
            return urls

        return img_urls

    def is_image(self, path: str) -> bool:
        """
        Check if url is image

        :param path: url
        :return: True if image
        """
        return path.lower().endswith(('.png', '.jpg', '.jpeg', '.tiff', '.bmp', '.gif', '.webp'))

    def encode_image(self, image_path: str) -> str:
        """
        Encode image to base64

        :param image_path: path to image
        :return: base64 encoded image
        """
        with open(image_path, "rb") as image_file:
            return base64.b64encode(image_file.read()).decode('utf-8')

    def reset_tokens(self):
        """Reset input tokens counter"""
        self.input_tokens = 0

    def reset(self):
        """Reset attachments, urls and input tokens"""
        self.attachments = {}
        self.urls = []
        self.input_tokens = 0

    def get_attachments(self) -> Dict[str, str]:
        """
        Get attachments

        :return: attachments dict
        """
        return self.attachments

    def get_urls(self) -> List[str]:
        """
        Get urls

        :return: urls list
        """
        return self.urls

    def get_used_tokens(self) -> int:
        """
        Get input tokens counter

        :return: input tokens
        """
        return self.input_tokens

    def append_images(self, ctx: CtxItem):
        """
        Append images content to context item

        :param ctx: context
        """
        images = self.get_attachments()  # dict -> key: id, value: path
        urls = self.get_urls()  # list

        # store sent images in ctx
        if len(images) > 0:
            ctx.images = self.window.core.filesystem.make_local_list(list(images.values()))
        if len(urls) > 0:
            ctx.images = urls
            ctx.urls = urls



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 22:00:00                  #
# ================================================== #

import hashlib
import os
from typing import Dict, Optional

from PySide6.QtGui import QAction, QIcon
from PySide6.QtWidgets import QFileDialog

from pygpt_net.tools.base import BaseTool
from pygpt_net.tools.text_editor.ui.dialogs import DialogSpawner
from pygpt_net.ui.widget.dialog.base import BaseDialog
from pygpt_net.utils import trans


class TextEditor(BaseTool):
    def __init__(self, *args, **kwargs):
        """
        Text editor

        :param window: Window instance
        """
        super(TextEditor, self).__init__(*args, **kwargs)
        self.id = "editor"
        self.width = 800
        self.height = 500
        self.instance_id = 0
        self.spawner = None

    def setup(self):
        """Setup tool"""
        self.spawner = DialogSpawner(self.window)

    def prepare_id(self, file: str):
        """
        Prepare unique id for file

        :param file: file name
        :return: unique id
        """
        return 'file_editor_' + hashlib.md5(file.encode('utf-8')).hexdigest()

    def open_file(
            self,
            id: str,
            auto_close: bool = True,
            force: bool = False,
            save: bool = False
    ):
        """
        Open text file dialog

        :param id: editor id
        :param auto_close: auto close current editor
        :param force: force open new instance
        :param save: save current file
        """
        if not force and self.window.core.filesystem.editor.is_changed(id) and auto_close:
            self.window.ui.dialogs.confirm(
                type='editor.changed.open',
                id=id,
                msg=trans("changed.confirm"),
            )
            return

        if save:
            self.save(id)

        path, _ = QFileDialog.getOpenFileName(
            self.window,
            trans("action.open"))
        if path:
            self.open(path, id, auto_close)

    def clear(
            self,
            id: str = None,
            force: bool = False,
            save: bool = False
    ):
        """
        Clear current instance

        :param id: editor id
        :param force: force open new instance
        :param save: save current file
        """
        if not force and self.window.core.filesystem.editor.is_changed(id):
            self.window.ui.dialogs.confirm(
                type='editor.changed.clear',
                id=id,
                msg=trans("changed.confirm"),
            )
            return

        if save:
            id = self.save(id)
            self.close(id)

        self.window.core.filesystem.editor.clear(id)

    def new(self):
        """Open new instance"""
        self.open()

    def open_batch(self, files: list):
        """
        Open multiple files in editor instances

        :param files: list of files to open
        """
        for file in files:
            self.open(file, auto_close=False, force=True)

    def open(
            self,
            file: Optional[str] = None,
            current_id: Optional[str] = None,
            auto_close: bool = True,
            force: bool = False):
        """
        Open/toggle file editor

        :param file: File to load
        :param current_id: editor id
        :param auto_close: auto close current editor
        :param force: force open new instance
        """
        # pre-check if file can be opened as text
        if file:
            try:
                with open(file, "r", encoding="utf-8") as f:
                    f.read()
            except Exception as e:
                self.window.ui.dialogs.alert(e)
                self.window.update_status("Error opening text file: {}".format(e))
                return

            # close current editor if different file
            id = self.prepare_id(file)
            if current_id and auto_close:
                if id != current_id:
                    self.close(current_id)
        else:
            # new instance id
            id = 'file_tmp_editor_' + str(self.instance_id)
            self.instance_id += 1

        self.window.ui.dialogs.open_instance(
            id,
            width=self.width,
            height=self.height,
        )

        # set file
        if file:
            title = os.path.basename(file)
            file_size = self.window.core.filesystem.sizeof_fmt(os.path.getsize(file))
            title += " - {}".format(file_size)
            self.window.ui.dialog[id].file = file
            self.window.core.filesystem.editor.load(id, file)  # load file to editor
            self.window.ui.dialog[id].setWindowTitle(title)  # set window title
            self.window.ui.editor[id].setFocus()  # set focus
            if not force:
                self.window.update_status("Loaded file: {}".format(os.path.basename(file)))  # set status
        else:
            self.window.core.filesystem.editor.clear(id)  # clear editor if no file

    def close(
            self,
            id: str,
            save: bool = False
    ):
        """
        Close file editor

        :param id: editor id
        :param save: save file before close
        """
        if save:
            id = self.save(id)
        self.window.ui.dialog[id].accept_close = True
        self.window.core.filesystem.editor.destroy(id)  # unregister from memory
        self.window.ui.dialogs.close(id)

    def save(self, id: str, close: bool = False) -> str:
        """
        Save content to current file

        :param id: editor id
        :param close: close editor after save
        :return: new editor id
        """
        file = self.window.ui.dialog[id].file
        if file:
            self.window.core.filesystem.editor.save(id)
            if close:
                self.close(id)
        else:
            id = self.save_as_file(id)
        return id

    def restore(
            self,
            id: str,
            force: bool = False,
            save: bool = False
    ):
        """
        Restore file content

        :param id: editor id
        :param force: force restore
        :param save: save file before restore
        """
        if not force and self.window.core.filesystem.editor.is_changed(id):
            self.window.ui.dialogs.confirm(
                type='editor.changed.restore',
                id=id,
                msg=trans("changed.confirm"),
            )
            return

        if save:
            self.save(id)

        file = self.window.ui.dialog[id].file
        if file:
            self.window.core.filesystem.editor.restore(id)
        else:
            self.window.ui.dialogs.alert("No file to restore")

    def save_as_file(self, id: str) -> str:
        """
        Save content to file

        :param id: editor id
        :return: new editor id
        """
        file_name = ""
        current = self.window.ui.dialog[id].file
        if current:
            file_name = os.path.basename(current)
        path, _ = QFileDialog.getSaveFileName(
            self.window,
            trans("action.save_as"),
            file_name,
        )
        if path:
            self.window.core.filesystem.editor.save(id, path)
        return id

    def setup_menu(self) -> Dict[str, QAction]:
        """
        Setup main menu

        :return dict with menu actions
        """
        actions = {}
        actions["text.editor"] = QAction(
            QIcon(":/icons/edit.svg"),
            trans("menu.tools.text.editor"),
            self.window,
            checkable=False,
        )
        actions["text.editor"].triggered.connect(
            lambda: self.open()
        )
        return actions

    def get_instance(
            self,
            type_id: str,
            dialog_id: Optional[str] = None
    ) -> Optional[BaseDialog]:
        """
        Spawn and return dialog instance

        :param type_id: dialog instance type ID
        :param dialog_id: dialog instance ID
        :return dialog instance
        """
        if type_id == "text_editor":
            return self.spawner.setup(dialog_id)

    def get_lang_mappings(self) -> Dict[str, Dict]:
        """
        Get language mappings

        :return: dict with language mappings
        """
        return {
            'menu.text': {
                'tools.text.editor': 'menu.tools.text.editor',
            }
        }



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.17 07:00:00                  #
# ================================================== #

from typing import List, Dict, Any, Optional

from PySide6.QtGui import QAction

from pygpt_net.core.types import (
    MODE_AGENT, MODE_AUDIO,
)
from pygpt_net.controller.plugins.presets import Presets
from pygpt_net.controller.plugins.settings import Settings
from pygpt_net.core.events import Event
from pygpt_net.item.ctx import CtxItem
from pygpt_net.utils import trans


class Plugins:
    def __init__(self, window=None):
        """
        Plugins controller

        :param window: Window instance
        """
        self.window = window
        self.settings = Settings(window)
        self.presets = Presets(window)
        self.enabled = {}
        self._ids = None
        self._ids_with_update = None
        self._suspend_updates = 0

    def _begin_batch(self):
        """Begin batch updates"""
        self._suspend_updates += 1

    def _end_batch(self):
        """End batch updates"""
        if self._suspend_updates > 0:
            self._suspend_updates -= 1
        if self._suspend_updates == 0:
            self.update_info()
            self.update()

    def setup(self):
        """Set up plugins"""
        self.setup_menu()
        self.setup_ui()

        try:
            self.window.core.plugins.clean_presets()
        except Exception as e:
            self.window.core.debug.error(e)

        self.presets.preset_to_current()
        self.reconfigure(silent=True)

    def reconfigure(self, silent: bool = False):
        """
        Reconfigure plugins

        :param silent: silent mode
        """
        self.setup_config(silent=silent)
        self.update()
        self.presets.update_menu()

    def setup_ui(self):
        """Set up plugins UI"""
        pm = self.window.core.plugins
        for pid in pm.get_ids():
            plugin = pm.get(pid)
            fn = getattr(plugin, "setup_ui", None)
            if callable(fn):
                fn()

        self.handle_types()

    def setup_menu(self):
        """Set up plugins menu"""
        pm = self.window.core.plugins
        ui_menu = self.window.ui.menu
        menu_plugins = ui_menu['plugins']
        for pid in pm.get_ids(sort=True):
            if pid in menu_plugins:
                continue
            name = pm.get_name(pid)
            tooltip = pm.get_desc(pid)
            act = QAction(name, self.window, checkable=True)
            act.triggered.connect(lambda checked=None, id=pid: self.toggle(id))
            act.setToolTip(tooltip)
            menu_plugins[pid] = act
            ui_menu['menu.plugins'].addAction(act)

    def setup_config(self, silent: bool = False):
        """
        Enable plugins from config

        :param silent: silent mode
        """
        pm = self.window.core.plugins
        cfg = self.window.core.config
        cfg_enabled = cfg.get('plugins_enabled')
        self._begin_batch()
        try:
            for pid in pm.get_ids():
                if pid in cfg_enabled:
                    if cfg.data['plugins_enabled'][pid]:
                        self.enable(pid)
                    else:
                        self.disable(pid, silent=silent)
                else:
                    self.disable(pid, silent=silent)
        finally:
            self._end_batch()

    def update(self):
        """Update plugins menu"""
        menu_plugins = self.window.ui.menu['plugins']
        for pid, action in menu_plugins.items():
            action.setChecked(self.enabled.get(pid, False))

        self.handle_types()
        self.window.controller.ui.mode.update()
        self.window.controller.ui.vision.update()

    def enable(self, id: str):
        """
        Enable plugin

        :param id: plugin id
        """
        pm = self.window.core.plugins
        if not pm.is_registered(id):
            return
        if self.enabled.get(id, False):
            return

        self.enabled[id] = True
        pm.enable(id)

        event = Event(Event.ENABLE, {'value': id})
        self.window.dispatch(event)

        if self.has_type(id, 'audio.input') or self.has_type(id, 'audio.output'):
            self.window.controller.audio.update()

        if self._suspend_updates == 0:
            self.update_info()
            self.update()

    def disable(self, id: str, silent: bool = False):
        """
        Disable plugin

        :param id: plugin id
        :param silent: silent mode
        """
        pm = self.window.core.plugins
        if not pm.is_registered(id):
            return
        if not self.enabled.get(id, False):
            return

        self.enabled[id] = False
        pm.disable(id)

        if not silent:
            event = Event(Event.DISABLE, {'value': id})
            self.window.dispatch(event, all=True)
            if self.has_type(id, 'audio.input') or self.has_type(id, 'audio.output'):
                self.window.controller.audio.update()

        if self._suspend_updates == 0:
            self.update_info()
            self.update()

    def is_enabled(self, id: str) -> bool:
        """
        Check if plugin is enabled

        :param id: plugin id
        :return: True if enabled
        """
        return self.window.core.plugins.is_registered(id) and self.enabled.get(id, False)

    def toggle(self, id: str):
        """
        Toggle plugin (from menu)

        :param id: plugin id
        """
        if self.window.core.plugins.is_registered(id):
            if self.is_enabled(id):
                self.disable(id)
                if id == "audio_output":
                    self.window.controller.audio.set_muted(True)
            else:
                self.enable(id)
                if id == "audio_output":
                    self.window.controller.audio.set_muted(False)

        self.window.controller.ui.update_tokens()
        self.window.controller.attachment.update()
        self.presets.save_current()

    def toggle_audio_output(self):
        """
        Toggle plugin (from icon, audio output only)
        """
        id = "audio_output"
        mode = self.window.core.config.get('mode')

        if mode == MODE_AUDIO:
            if not self.window.controller.audio.is_muted():
                self.window.controller.audio.set_muted(True)
            else:
                self.window.controller.audio.set_muted(False)
            return

        if self.window.core.plugins.is_registered(id):
            if self.is_enabled(id):
                self.disable(id)
            else:
                self.enable(id)

        self.presets.save_current()

    def set_by_tab(self, idx: int):
        """
        Set current plugin by tab index

        :param idx: tab index
        """
        pm = self.window.core.plugins
        plugin_idx = 0
        for pid in pm.get_ids(sort=True):
            if pm.has_options(pid):
                if plugin_idx == idx:
                    self.settings.current_plugin = pid
                    break
                plugin_idx += 1
        current = self.window.ui.models['plugin.list'].index(idx, 0)
        self.window.ui.nodes['plugin.list'].setCurrentIndex(current)

    def get_tab_idx(self, plugin_id: str) -> Optional[int]:
        """
        Get plugin tab index

        :param plugin_id: plugin id
        :return: tab index or None if not found
        """
        pm = self.window.core.plugins
        for i, pid in enumerate(pm.get_ids(sort=True)):
            if pid == plugin_id:
                return i
        return None

    def unregister(self, id: str):
        """
        Unregister plugin

        :param id: plugin id
        """
        self.window.core.plugins.unregister(id)
        self.enabled.pop(id, None)

    def destroy(self):
        """Destroy plugins workers"""
        event = Event(Event.FORCE_STOP, {})
        self.window.dispatch(event)

        pm = self.window.core.plugins
        for pid in pm.get_ids():
            plugin = pm.get(pid)
            fn = getattr(pm, "destroy", None)
            if callable(fn):
                try:
                    pm.destroy(pid)
                except AttributeError:
                    pass

    def has_type(self, id: str, type: str):
        """
        Check if plugin has type

        :param id: plugin ID
        :param type: type to check
        :return: True if has type
        """
        pm = self.window.core.plugins
        if not pm.is_registered(id):
            return False
        return type in pm.get(id).type

    def is_type_enabled(self, type: str) -> bool:
        """
        Check if plugin type is enabled

        :param type: plugin type
        :return: True if enabled
        """
        pm = self.window.core.plugins
        return any((type in pm.get(pid).type) and self.is_enabled(pid) for pid in pm.get_ids())

    def handle_types(self):
        """Handle plugin type"""
        pm = self.window.core.plugins
        for t in pm.allowed_types:
            enabled = self.is_type_enabled(t)
            if t == 'audio.input':
                self.window.controller.audio.handle_audio_input(enabled)
            elif t == 'audio.output':
                self.window.controller.audio.handle_audio_output(enabled)
            elif t == 'schedule':
                if enabled:
                    self.window.ui.plugin_addon['schedule'].setVisible(True)
                    data = {'name': 'scheduled_tasks_count', 'value': 0}
                    event = Event(Event.PLUGIN_OPTION_GET, data)
                    self.window.dispatch(event)
                    num = event.data.get('value', 0)
                    self.window.ui.tray.update_schedule_tasks(num)
                else:
                    self.window.ui.plugin_addon['schedule'].setVisible(False)
                    self.window.ui.tray.hide_schedule_menu()

    def on_update(self):
        """Called on update"""
        pm = self.window.core.plugins
        for pid in pm.get_ids():
            if self.is_enabled(pid):
                fn = getattr(pm.get(pid), "on_update", None)
                if callable(fn):
                    fn()

    def on_post_update(self):
        """Called on post update"""
        pm = self.window.core.plugins
        if self._ids is None:
            self._ids = pm.get_ids()
        if self._ids_with_update is None:
            self._ids_with_update = [pid for pid in self._ids if hasattr(self.window.core.plugins.get(pid), "on_post_update")]
        if len(self._ids_with_update) == 0:
            return
        for pid in self._ids_with_update:
            if self.is_enabled(pid):
                fn = pm.get(pid).on_post_update
                if callable(fn):
                    fn()

    def update_info(self):
        """Update plugins info"""
        pm = self.window.core.plugins
        enabled_names = []
        c = 0
        for pid in pm.get_ids():
            if self.is_enabled(pid):
                c += 1
                enabled_names.append(pm.get(pid).name)

        tooltip = " + ".join(enabled_names)
        count_str = f"+ {c} {trans('chatbox.plugins')}" if c > 0 else ""
        self.window.ui.nodes['chat.plugins'].setText(count_str)
        self.window.ui.nodes['chat.plugins'].setToolTip(tooltip)

    def _apply_cmds_common(
            self,
            event_type: str,
            ctx: CtxItem,
            cmds: List[Dict[str, Any]],
            all: bool = False,
            execute_only: bool = False
    ) -> Optional[List[Any]]:
        """
        Common method to apply commands

        This method is used for both inline and non-inline commands.

        :param event_type: name of the event type, either Event.CMD_EXECUTE or Event.CMD_INLINE
        :param ctx: CtxItem
        :param cmds: commands list, each command is a dictionary with keys like "cmd", "args", etc.
        :param all: True to apply all commands, False to apply only enabled commands
        :param execute_only: True to execute commands only, without any additional event
        :return: results: results of the command execution, if any (ctx.results)
        """
        commands = self.window.core.command.from_commands(cmds)
        if len(commands) == 0:
            return

        event = Event(event_type, {'commands': commands})
        mode = self.window.core.config.get('mode')
        self.log("Executing plugin commands..." if event_type == Event.CMD_EXECUTE else "Executing inline plugin commands...")
        change_status = True
        if mode == MODE_AGENT and len(cmds) == 1 and cmds[0].get("cmd") == "goal_update":
            change_status = False
        wait_str = trans('status.cmd.wait')
        if change_status:
            self.window.update_status(wait_str)

        ctx.results = []
        event.ctx = ctx
        if event_type == Event.CMD_EXECUTE:
            self.window.controller.command.dispatch(event, all=all, execute_only=execute_only)
        else:
            self.window.controller.command.dispatch(event)

        current = self.window.ui.get_status()
        if current == wait_str:
            self.window.update_status("")
        return ctx.results

    def apply_cmds(
            self,
            ctx: CtxItem,
            cmds: List[Dict[str, Any]],
            all: bool = False,
            execute_only: bool = False
    ) -> Optional[List[Any]]:
        """
        Apply commands (CMD_EXECUTE event only)

        :param ctx: CtxItem
        :param cmds: commands list
        :param all: True to apply all commands, False to apply only enabled commands
        :param execute_only: True to execute commands only, without any additional event
        :return: results: results of the command execution, if any (ctx.results)
        """
        return self._apply_cmds_common(Event.CMD_EXECUTE, ctx, cmds, all=all, execute_only=execute_only)

    def apply_cmds_all(
            self,
            ctx: CtxItem,
            cmds: List[Dict[str, Any]]
    ) -> Optional[List[Any]]:
        """
        Apply all commands (inline or not)

        :param ctx: CtxItem
        :param cmds: commands list
        :return: results: results of the command execution, if any (ctx.results)
        """
        if self.window.core.config.get("cmd"):
            return self.apply_cmds(ctx, cmds)
        else:
            return self.apply_cmds_inline(ctx, cmds)


    def apply_cmds_inline(
            self,
            ctx: CtxItem,
            cmds: List[Dict[str, Any]]
    ) -> Optional[List[Any]]:
        """
        Apply inline commands

        :param ctx: CtxItem
        :param cmds: commands list
        :return: results: results of the command execution, if any (ctx.results)
        """
        return self._apply_cmds_common(Event.CMD_INLINE, ctx, cmds)

    def reload(self):
        """Reload plugins"""
        self.window.core.plugins.reload_all()
        self.setup()
        self.settings.init()
        self.update()

    def save_all(self):
        """Save plugin settings"""
        pm = self.window.core.plugins
        cfg_plugins = self.window.core.config.data['plugins']

        for pid, plugin in pm.plugins.items():
            plugin.setup()
            if pid not in cfg_plugins:
                cfg_plugins[pid] = {}
            dest = cfg_plugins[pid]
            for key, opt in plugin.options.items():
                dest[key] = opt['value']

        for key in list(cfg_plugins.keys()):
            if key not in pm.plugins:
                cfg_plugins.pop(key)

        self.window.controller.plugins.presets.save_current()
        self.window.core.config.save()

    def log(self, data: Any):
        """
        Log data to debug

        :param data: Data to log
        """
        self.window.core.debug.info(data)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.12.27 17:00:00                  #
# ================================================== #

import datetime
import os
import shutil
import struct
import sys
from typing import Union

from PySide6.QtCore import Qt, QModelIndex, QDir, QObject, QEvent, QUrl, QPoint, QMimeData, QTimer, QRect, QItemSelectionModel
from PySide6.QtGui import QAction, QIcon, QCursor, QResizeEvent, QGuiApplication, QKeySequence, QShortcut, QClipboard, QDrag
from PySide6.QtWidgets import QTreeView, QMenu, QWidget, QVBoxLayout, QFileSystemModel, QLabel, QHBoxLayout, \
    QPushButton, QSizePolicy, QAbstractItemView, QFrame

from pygpt_net.core.tabs.tab import Tab
from pygpt_net.ui.widget.element.button import ContextMenuButton
from pygpt_net.ui.widget.element.labels import HelpLabel
from pygpt_net.utils import trans


class MultiDragTreeView(QTreeView):
    """
    QTreeView with improved multi-selection UX:
    - When multiple rows are already selected and user presses left mouse (no modifiers),
      a short press-release clears selection (global single-click to deselect),
      but moving the mouse beyond the drag threshold starts a drag containing the whole selection
      instead of altering selection.
    - This avoids accidental selection changes when the intent was to drag many items.
    - Shift-range selection anchor is kept stable to avoid selecting the entire directory accidentally.
    """
    def __init__(self, parent=None):
        super().__init__(parent)
        self._md_pressed = False
        self._md_drag_started = False
        self._md_maybe_clear = False
        self._md_press_pos = QPoint()
        self._md_press_index = QModelIndex()
        self._sel_anchor_index = QModelIndex()

    def _selected_count(self) -> int:
        try:
            return len(self.selectionModel().selectedRows(0))
        except Exception:
            return 0

    def _make_urls_from_selection(self):
        urls = []
        try:
            model = self.model()
            rows = self.selectionModel().selectedRows(0)
            seen = set()
            for idx in rows:
                p = model.filePath(idx)
                if p and p not in seen:
                    seen.add(p)
                    urls.append(QUrl.fromLocalFile(p))
        except Exception:
            pass
        return urls

    def _start_multi_drag(self):
        urls = self._make_urls_from_selection()
        if not urls:
            return
        md = QMimeData()
        try:
            parts = []
            for u in urls:
                try:
                    parts.append(u.toString(QUrl.FullyEncoded))
                except Exception:
                    parts.append(u.toString())
            md.setData("text/uri-list", ("\r\n".join(parts) + "\r\n").encode("utf-8"))
        except Exception:
            pass
        md.setUrls(urls)

        drag = QDrag(self)
        drag.setMimeData(md)
        drag.exec(Qt.MoveAction | Qt.CopyAction, Qt.MoveAction)

    def _drag_threshold(self) -> int:
        """
        Return platform drag threshold using style hints when available.
        Compatible with Qt 6 where static startDragDistance() is not exposed on QGuiApplication.
        """
        try:
            hints = QGuiApplication.styleHints()
            if hints is not None:
                getter = getattr(hints, "startDragDistance", None)
                if callable(getter):
                    return int(getter())
                val = getattr(hints, "startDragDistance", 0)
                if isinstance(val, int) and val > 0:
                    return val
        except Exception:
            pass
        try:
            from PySide6.QtWidgets import QApplication
            return int(QApplication.startDragDistance())
        except Exception:
            pass
        return 10

    def _event_point(self, event) -> QPoint:
        """Return mouse point for Qt versions exposing either .position() or .pos()."""
        try:
            return event.position().toPoint()
        except Exception:
            try:
                return event.pos()
            except Exception:
                return QPoint()

    def _row_index_at(self, pos: QPoint) -> QModelIndex:
        """
        Return a model index for the row under 'pos' forced to column 0,
        so row-based selection is consistent no matter which column is clicked.
        """
        try:
            idx = self.indexAt(pos)
            if idx.isValid():
                try:
                    return idx.siblingAtColumn(0)
                except Exception:
                    return self.model().index(idx.row(), 0, idx.parent())
        except Exception:
            pass
        return QModelIndex()

    def mousePressEvent(self, event):
        pos = self._event_point(event)
        if event.button() == Qt.LeftButton:
            ctrl = bool(event.modifiers() & Qt.ControlModifier)
            shift = bool(event.modifiers() & Qt.ShiftModifier)

            pressed_idx = self._row_index_at(pos)
            self._md_press_index = pressed_idx

            if shift:
                try:
                    sm = self.selectionModel()
                except Exception:
                    sm = None
                if sm is not None:
                    anchor = self._sel_anchor_index if self._sel_anchor_index.isValid() else sm.currentIndex()
                    if not (anchor and anchor.isValid()):
                        anchor = pressed_idx
                        self._sel_anchor_index = pressed_idx
                    try:
                        sm.setCurrentIndex(anchor, QItemSelectionModel.NoUpdate)
                    except Exception:
                        pass
                super().mousePressEvent(event)
                return

            if not ctrl and not shift:
                if pressed_idx.isValid():
                    self._sel_anchor_index = pressed_idx
                    try:
                        sm = self.selectionModel()
                        if sm is not None:
                            sm.setCurrentIndex(pressed_idx, QItemSelectionModel.NoUpdate)
                    except Exception:
                        pass

                if self._selected_count() > 1:
                    self._md_pressed = True
                    self._md_drag_started = False
                    self._md_maybe_clear = True
                    self._md_press_pos = pos
                    self.setFocus(Qt.MouseFocusReason)
                    event.accept()
                    return

        super().mousePressEvent(event)

    def mouseMoveEvent(self, event):
        if self._md_pressed and (event.buttons() & Qt.LeftButton):
            pos = self._event_point(event)
            if (pos - self._md_press_pos).manhattanLength() >= self._drag_threshold():
                self._md_maybe_clear = False
                self._md_drag_started = True
                self._md_pressed = False
                self._start_multi_drag()
                event.accept()
                return
        super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event):
        if self._md_pressed and event.button() == Qt.LeftButton:
            if self._md_maybe_clear and not self._md_drag_started:
                try:
                    sm = self.selectionModel()
                except Exception:
                    sm = None
                if sm is not None:
                    sm.clearSelection()
                if self._md_press_index.isValid():
                    try:
                        self.setCurrentIndex(self._md_press_index)
                    except Exception:
                        pass
                    self._sel_anchor_index = self._md_press_index
                else:
                    self.setCurrentIndex(QModelIndex())

                event.accept()
                self._md_pressed = False
                return
        self._md_pressed = False
        super().mouseReleaseEvent(event)


class ExplorerDropHandler(QObject):
    """
    Drag & drop handler for FileExplorer (uploads and internal moves).
    - Accepts local file and directory URLs.
    - Target directory rules based on mouse position:
      * hovering directory (not in left gutter) -> into that directory
      * hovering a row in its left gutter       -> into parent directory (one level up)
      * between two rows                        -> into their common parent (or explorer root if top-level)
      * empty area                              -> explorer root
    - Internal drags result in move; external drags result in copy/upload.
    - Provides manual auto-scroll during drag.
    - Visuals:
      * highlight rectangle when targeting a directory row
      * horizontal indicator line snapped between rows for "between" drops
    """
    def __init__(self, explorer):
        super().__init__(explorer)
        self.explorer = explorer
        self.view = explorer.treeView

        try:
            self.view.setAcceptDrops(True)
        except Exception:
            pass
        vp = self.view.viewport()
        if vp is not None:
            try:
                vp.setAcceptDrops(True)
            except Exception:
                pass
            vp.installEventFilter(self)
        self.view.installEventFilter(self)

        self._indicator = QFrame(self.view.viewport())
        self._indicator.setObjectName("drop-indicator-line")
        self._indicator.setFrameShape(QFrame.NoFrame)
        self._indicator.setStyleSheet("#drop-indicator-line { background-color: rgba(40,120,255,0.95); }")
        self._indicator.hide()
        self._indicator_height = 2

        self._dir_highlight = QFrame(self.view.viewport())
        self._dir_highlight.setObjectName("drop-dir-highlight")
        self._dir_highlight.setFrameShape(QFrame.NoFrame)
        self._dir_highlight.setStyleSheet(
            "#drop-dir-highlight { border: 2px solid rgba(40,120,255,0.95); border-radius: 3px; "
            "background-color: rgba(40,120,255,0.10); }"
        )
        self._dir_highlight.hide()

        self._scroll_margin = 28
        self._scroll_speed_max = 12
        self._last_pos = None
        self._auto_timer = QTimer(self)
        self._auto_timer.setInterval(20)
        self._auto_timer.timeout.connect(self._on_auto_scroll)

        self._left_gutter_extra = 28

    def _mime_has_local_urls(self, md) -> bool:
        try:
            if md and md.hasUrls():
                for url in md.urls():
                    if url.isLocalFile():
                        return True
        except Exception:
            pass
        return False

    def _local_paths_from_mime(self, md) -> list:
        out = []
        try:
            if not (md and md.hasUrls()):
                return out
            for url in md.urls():
                try:
                    if url.isLocalFile():
                        p = url.toLocalFile()
                        if p:
                            out.append(p)
                except Exception:
                    continue
        except Exception:
            pass
        return out

    def _nearest_row_index(self, pos: QPoint):
        idx = self.view.indexAt(pos)
        if idx.isValid():
            return idx

        vp = self.view.viewport()
        h = vp.height()
        for d in range(1, 65):
            y_up = pos.y() - d
            if y_up >= 0:
                idx_up = self.view.indexAt(QPoint(pos.x(), y_up))
                if idx_up.isValid():
                    return idx_up
            y_dn = pos.y() + d
            if y_dn < h:
                idx_dn = self.view.indexAt(QPoint(pos.x(), y_dn))
                if idx_dn.isValid():
                    return idx_dn
        return QModelIndex()

    def _indices_above_below(self, pos: QPoint):
        vp = self.view.viewport()
        h = vp.height()
        above = QModelIndex()
        below = QModelIndex()

        for d in range(0, 80):
            y_up = pos.y() - d
            if y_up < 0:
                break
            idx_up = self.view.indexAt(QPoint(max(0, pos.x()), y_up))
            if idx_up.isValid():
                above = idx_up
                break

        for d in range(0, 80):
            y_dn = pos.y() + d
            if y_dn >= h:
                break
            idx_dn = self.view.indexAt(QPoint(max(0, pos.x()), y_dn))
            if idx_dn.isValid():
                below = idx_dn
                break

        return above, below

    def _gap_parent_index(self, pos: QPoint) -> QModelIndex:
        above, below = self._indices_above_below(pos)
        if above.isValid() and below.isValid():
            pa = above.parent()
            pb = below.parent()
            if pa == pb:
                return pa
            if pa.isValid():
                return pa
            if pb.isValid():
                return pb
        elif above.isValid():
            return above.parent()
        elif below.isValid():
            return below.parent()
        return QModelIndex()

    def _is_left_gutter(self, pos: QPoint, idx: QModelIndex) -> bool:
        if not idx.isValid():
            return False
        rect = self.view.visualRect(idx)
        return pos.x() < rect.left() + self._left_gutter_extra

    def _snap_line_y(self, pos: QPoint) -> int:
        vp = self.view.viewport()
        y = max(0, min(pos.y(), vp.height() - 1))
        idx = self._nearest_row_index(pos)
        if idx.isValid():
            rect = self.view.visualRect(idx)
            if y < rect.center().y():
                return rect.top()
            return rect.bottom()
        return y

    def _calc_context(self, pos: QPoint):
        vp = self.view.viewport()
        if vp is None:
            return {'type': 'empty'}

        idx = self.view.indexAt(pos)
        if idx.isValid():
            if self._is_left_gutter(pos, idx):
                rect = self.view.visualRect(idx)
                line_y = rect.top() if pos.y() < rect.center().y() else rect.bottom()
                return {'type': 'into_parent', 'idx': idx, 'parent_idx': idx.parent(), 'line_y': line_y}

            path = self.explorer.model.filePath(idx)
            if os.path.isdir(path):
                return {'type': 'into_dir', 'idx': idx}
            rect = self.view.visualRect(idx)
            line_y = rect.top() if pos.y() < rect.center().y() else rect.bottom()
            return {'type': 'into_parent', 'idx': idx, 'parent_idx': idx.parent(), 'line_y': line_y}

        parent_idx = self._gap_parent_index(pos)
        return {'type': 'gap_between', 'parent_idx': parent_idx, 'line_y': self._snap_line_y(pos)}

    def _update_visuals(self, ctx):
        self._indicator.hide()
        self._dir_highlight.hide()

        if not isinstance(ctx, dict):
            return

        if ctx.get('type') == 'into_dir' and ctx.get('idx', QModelIndex()).isValid():
            rect = self.view.visualRect(ctx['idx'])
            self._dir_highlight.setGeometry(QRect(0, rect.top(), self.view.viewport().width(), rect.height()))
            self._dir_highlight.show()
            return

        if ctx.get('type') in ('gap_between', 'into_parent'):
            y = ctx.get('line_y', None)
            if y is None:
                return
            self._indicator.setGeometry(QRect(0, y, self.view.viewport().width(), self._indicator_height))
            self._indicator.show()
            return

    def _on_auto_scroll(self):
        if self._last_pos is None:
            return
        vp = self.view.viewport()
        y = self._last_pos.y()
        h = vp.height()
        vbar = self.view.verticalScrollBar()
        delta = 0

        if y < self._scroll_margin:
            strength = max(0.0, (self._scroll_margin - y) / self._scroll_margin)
            delta = -max(1, int(strength * self._scroll_speed_max))
        elif y > h - self._scroll_margin:
            strength = max(0.0, (y - (h - self._scroll_margin)) / self._scroll_margin)
            delta = max(1, int(strength * self._scroll_speed_max))

        if delta != 0 and vbar is not None:
            vbar.setValue(vbar.value() + delta)

    def _target_dir_from_context(self, ctx) -> str:
        t = ctx.get('type')
        if t == 'into_dir':
            idx = ctx.get('idx', QModelIndex())
            if idx.isValid():
                path = self.explorer.model.filePath(idx)
                if os.path.isdir(path):
                    return path
        elif t in ('gap_between', 'into_parent'):
            parent_idx = ctx.get('parent_idx', QModelIndex())
            if parent_idx.isValid():
                return self.explorer.model.filePath(parent_idx)
            return self.explorer.directory
        return self.explorer.directory

    def _is_internal_drag(self, event) -> bool:
        try:
            src = event.source()
            return src is not None and (src is self.view or src is self.view.viewport())
        except Exception:
            return False

    def eventFilter(self, obj, event):
        et = event.type()

        if et == QEvent.DragEnter:
            md = getattr(event, 'mimeData', lambda: None)()
            if self._mime_has_local_urls(md):
                try:
                    if self._is_internal_drag(event):
                        event.setDropAction(Qt.MoveAction)
                    else:
                        event.setDropAction(Qt.CopyAction)
                    event.acceptProposedAction()
                except Exception:
                    event.accept()
                try:
                    self._last_pos = event.position().toPoint()
                except Exception:
                    try:
                        self._last_pos = event.pos()
                    except Exception:
                        self._last_pos = None
                self._auto_timer.start()

                pos = self._last_pos or QPoint(0, 0)
                ctx = self._calc_context(pos)
                self._update_visuals(ctx)
                return True
            return False

        if et == QEvent.DragMove:
            md = getattr(event, 'mimeData', lambda: None)()
            if self._mime_has_local_urls(md):
                try:
                    if self._is_internal_drag(event):
                        event.setDropAction(Qt.MoveAction)
                    else:
                        event.setDropAction(Qt.CopyAction)
                    event.acceptProposedAction()
                except Exception:
                    event.accept()

                try:
                    self._last_pos = event.position().toPoint()
                except Exception:
                    try:
                        self._last_pos = event.pos()
                    except Exception:
                        self._last_pos = None

                pos = self._last_pos or QPoint(0, 0)
                ctx = self._calc_context(pos)
                self._update_visuals(ctx)
                return True
            return False

        if et in (QEvent.DragLeave, QEvent.Leave):
            self._auto_timer.stop()
            self._indicator.hide()
            self._dir_highlight.hide()
            self._last_pos = None
            return False

        if et == QEvent.Drop:
            self._auto_timer.stop()
            self._indicator.hide()
            self._dir_highlight.hide()

            md = getattr(event, 'mimeData', lambda: None)()
            if not self._mime_has_local_urls(md):
                return False

            try:
                pos = event.position().toPoint()
            except Exception:
                pos = event.pos() if hasattr(event, "pos") else QPoint()

            ctx = self._calc_context(pos)
            target_dir = self._target_dir_from_context(ctx)

            paths = self._local_paths_from_mime(md)
            is_internal = self._is_internal_drag(event)

            dest_paths = []
            try:
                if is_internal:
                    dest_paths = self.explorer._move_paths(paths, target_dir)
                else:
                    try:
                        self.explorer.window.controller.files.upload_paths(paths, target_dir)
                        dest_paths = [os.path.join(target_dir, os.path.basename(p.rstrip(os.sep))) for p in paths]
                    except Exception:
                        dest_paths = self.explorer._copy_paths(paths, target_dir)
                if os.path.isdir(target_dir):
                    self.explorer._expand_dir(target_dir, center=False)
                if dest_paths:
                    self.explorer._reveal_paths(dest_paths, select_first=True)
            except Exception as e:
                try:
                    self.explorer.window.core.debug.log(e)
                except Exception:
                    pass

            try:
                event.setDropAction(Qt.MoveAction if is_internal else Qt.CopyAction)
                event.acceptProposedAction()
            except Exception:
                event.accept()
            return True

        return False


class FileExplorer(QWidget):
    def __init__(self, window, directory, index_data):
        """
        File explorer widget

        :param window: Window instance
        :param directory: directory to explore
        :param index_data: index data
        """
        super().__init__()

        self.window = window
        self.owner = None
        self.index_data = index_data
        self.directory = directory
        self.model = IndexedFileSystemModel(self.window, self.index_data)
        self.model.setRootPath(self.directory)
        self.model.setFilter(self.model.filter() | QDir.Hidden)
        self.treeView = MultiDragTreeView()
        self.treeView.setModel(self.model)
        self.treeView.setRootIndex(self.model.index(self.directory))
        self.treeView.setUniformRowHeights(True)
        self.setProperty('class', 'file-explorer')

        # Multi-selection support via Ctrl/Shift and row-based selection
        try:
            self.treeView.setSelectionMode(QAbstractItemView.ExtendedSelection)
            self.treeView.setSelectionBehavior(QAbstractItemView.SelectRows)
        except Exception:
            pass

        header = QHBoxLayout()

        self.btn_open = QPushButton(trans('action.open'))
        self.btn_open.setMaximumHeight(40)
        self.btn_open.clicked.connect(
                lambda: self.action_open(self.directory)
        )
        self.btn_open.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)

        self.btn_upload = QPushButton(trans('files.local.upload'))
        self.btn_upload.setMaximumHeight(40)
        self.btn_upload.clicked.connect(self.window.controller.files.upload_local)
        self.btn_upload.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)

        self.btn_idx = ContextMenuButton(trans('idx.btn.index_all'), self)
        self.btn_idx.action = self.idx_context_menu
        self.btn_idx.setMaximumHeight(40)
        self.btn_idx.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)

        self.btn_clear = ContextMenuButton(trans('idx.btn.clear'), self)
        self.btn_clear.action = self.clear_context_menu
        self.btn_clear.setMaximumHeight(40)
        self.btn_clear.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)

        self.btn_tool = QPushButton(QIcon(":/icons/db.svg"), "")
        self.btn_tool.setMaximumHeight(40)
        self.btn_tool.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
        self.btn_tool.clicked.connect(
            lambda: self.window.tools.get("indexer").toggle()
        )

        self.path_label = QLabel(self.directory)
        self.path_label.setMaximumHeight(40)
        self.path_label.setAlignment(Qt.AlignCenter | Qt.AlignVCenter)

        header.addWidget(self.btn_open)
        header.addWidget(self.btn_upload)
        header.addStretch()
        header.addWidget(self.path_label)
        header.addStretch()

        header.addWidget(self.btn_tool)
        header.addWidget(self.btn_idx)
        header.addWidget(self.btn_clear)
        self.layout = QVBoxLayout()

        self.window.ui.nodes['tip.output.tab.files'] = HelpLabel(trans('tip.output.tab.files'), self.window)

        self.layout.addWidget(self.treeView)
        self.layout.addWidget(self.window.ui.nodes['tip.output.tab.files'])
        self.layout.addLayout(header)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self.layout)

        self.treeView.setContextMenuPolicy(Qt.CustomContextMenu)
        self.treeView.customContextMenuRequested.connect(self.openContextMenu)
        self.treeView.setColumnWidth(0, int(self.width() / 2))

        self.header = self.treeView.header()
        self.header.setStretchLastSection(True)
        self.header.setContentsMargins(0, 0, 0, 0)

        # Persisted column widths across model refreshes/layout changes
        self._saved_column_widths = {}
        self._is_restoring_columns = False

        # Re-apply widths when user resizes columns or model refreshes
        try:
            self.header.sectionResized.connect(self._on_header_section_resized)
            self.model.modelAboutToBeReset.connect(self._on_model_about_to_reset)
            self.model.modelReset.connect(self._on_model_reset)
            self.model.layoutAboutToBeChanged.connect(self._on_layout_about_to_change)
            self.model.layoutChanged.connect(self._on_layout_changed)
            self.model.directoryLoaded.connect(self._on_model_directory_loaded)
        except Exception:
            pass

        self.column_proportion = 0.3
        self.adjustColumnWidths()  # initial layout

        self.header.setStyleSheet("""
           QHeaderView::section {
               text-align: center;
               vertical-align: middle;
           }
       """)
        self.tab = None
        self.installEventFilter(self)
        try:
            self.treeView.viewport().installEventFilter(self)
        except Exception:
            pass

        self._icons = {
            'open': QIcon(":/icons/view.svg"),
            'open_dir': QIcon(":/icons/folder_filled.svg"),
            'download': QIcon(":/icons/download.svg"),
            'rename': QIcon(":/icons/edit.svg"),
            'duplicate': QIcon(":/icons/stack.svg"),
            'touch': QIcon(":/icons/add.svg"),
            'mkdir': QIcon(":/icons/add_folder.svg"),
            'refresh': QIcon(":/icons/reload.svg"),
            'upload': QIcon(":/icons/upload.svg"),
            'delete': QIcon(":/icons/delete.svg"),
            'attachment': QIcon(":/icons/attachment.svg"),
            'copy': QIcon(":/icons/copy.svg"),
            'cut': QIcon(":/icons/cut.svg"),
            'paste': QIcon(":/icons/paste.svg"),
            'read': QIcon(":/icons/view.svg"),
            'db': QIcon(":/icons/db.svg"),
            'pack': QIcon(":/icons/upload.svg"),
            'unpack': QIcon(":/icons/download.svg"),
        }

        try:
            self.treeView.setDragEnabled(True)
            self.treeView.setAcceptDrops(True)
            self.treeView.setDropIndicatorShown(False)
            self.treeView.setDragDropMode(QAbstractItemView.DragDrop)
            self.treeView.setDefaultDropAction(Qt.MoveAction)
            self.treeView.setAutoScroll(False)
        except Exception:
            pass

        self._cb_paths = []
        self._cb_mode = None

        try:
            sc_copy = QShortcut(QKeySequence.Copy, self.treeView, context=Qt.WidgetWithChildrenShortcut)
            sc_copy.activated.connect(self.action_copy_selection)
            sc_cut = QShortcut(QKeySequence.Cut, self.treeView, context=Qt.WidgetWithChildrenShortcut)
            sc_cut.activated.connect(self.action_cut_selection)
            sc_paste = QShortcut(QKeySequence.Paste, self.treeView, context=Qt.WidgetWithChildrenShortcut)
            sc_paste.activated.connect(self.action_paste_into_current)
        except Exception:
            pass

        self._dnd_handler = ExplorerDropHandler(self)

    def _on_header_section_resized(self, logical_index: int, old: int, new: int):
        """Track user-driven column width changes."""
        if self._is_restoring_columns:
            return
        try:
            self._saved_column_widths[logical_index] = int(new)
        except Exception:
            pass

    def _on_model_about_to_reset(self):
        """Save current widths before model reset."""
        self._save_current_column_widths()

    def _on_model_reset(self):
        """Restore widths right after model reset."""
        self._schedule_restore_columns()

    def _on_layout_about_to_change(self):
        """Save widths before layout changes."""
        self._save_current_column_widths()

    def _on_layout_changed(self):
        """Restore widths after layout changes."""
        self._schedule_restore_columns()

    def _on_model_directory_loaded(self, path: str):
        """Ensure widths are re-applied when a directory finishes loading."""
        self._schedule_restore_columns()

    def _save_current_column_widths(self):
        """Persist current column widths."""
        try:
            count = self.model.columnCount()
            for i in range(count):
                w = self.treeView.columnWidth(i)
                if w > 0:
                    self._saved_column_widths[i] = int(w)
        except Exception:
            pass

    def _restore_columns_now(self):
        """Best-effort restoration of column widths with proportional fallback."""
        try:
            self._is_restoring_columns = True
            col_count = self.model.columnCount()
            self.adjustColumnWidths()  # apply proportional baseline
            if self._saved_column_widths:
                for i, w in list(self._saved_column_widths.items()):
                    if 0 <= i < col_count and w > 0:
                        try:
                            self.treeView.setColumnWidth(i, int(w))
                        except Exception:
                            pass
            try:
                self.header.setStretchLastSection(True)
            except Exception:
                pass
        finally:
            self._is_restoring_columns = False

    def _schedule_restore_columns(self, delay_ms: int = 0):
        """Defer restoration to allow view/header to settle after reset."""
        QTimer.singleShot(max(0, delay_ms), self._restore_columns_now)

    def eventFilter(self, source, event):
        """
        Focus event filter

        :param source: source
        :param event: event
        """
        if event.type() == event.Type.FocusIn:
            if self.tab is not None:
                col_idx = self.tab.column_idx
                self.window.controller.ui.tabs.on_column_focus(col_idx)
        return super().eventFilter(source, event)

    def set_tab(self, tab: Tab):
        """
        Set tab

        :param tab: Tab
        """
        self.tab = tab

    def setOwner(self, owner: Tab):
        """
        Set tab parent (owner)

        :param owner: parent tab instance
        """
        self.owner = owner

    def getOwner(self) -> Tab:
        """
        Get tab parent (owner)

        :return: parent tab instance
        """
        return self.owner

    def update_view(self):
        """Update explorer view keeping column widths intact."""
        self._save_current_column_widths()
        self.model.beginResetModel()
        self.model.setRootPath(self.directory)
        self.model.endResetModel()
        self.treeView.setRootIndex(self.model.index(self.directory))
        self._schedule_restore_columns()

    def idx_context_menu(self, parent, pos):
        """
        Index all btn context menu

        :param parent: parent widget
        :param pos: mouse  position
        """
        menu = QMenu(self)
        idx_list = self.window.core.config.get('llama.idx.list')
        if len(idx_list) > 0:
            for idx in idx_list:
                id = idx['id']
                name = f"{idx['name']} ({idx['id']})"
                action = menu.addAction(f"IDX: {name}")
                action.triggered.connect(
                    lambda checked=False,
                           id=id: self.window.controller.idx.indexer.index_all_files(id)
                )
        menu.exec(parent.mapToGlobal(pos))

    def clear_context_menu(self, parent, pos):
        """
        Clear btn context menu

        :param parent: parent widget
        :param pos: mouse position
        """
        menu = QMenu(self)
        idx_list = self.window.core.config.get('llama.idx.list')
        if len(idx_list) > 0:
            for idx in idx_list:
                id = idx['id']
                name = f"{idx['name']} ({idx['id']})"
                action = menu.addAction(f"IDX: {name}")
                action.triggered.connect(
                    lambda checked=False,
                           id=id: self.window.controller.idx.indexer.clear(id)
                )
        menu.exec(parent.mapToGlobal(pos))

    def adjustColumnWidths(self):
        """Adjust column widths and persist them."""
        total_width = self.treeView.width()
        col_count = self.model.columnCount()
        first_column_width = int(total_width * self.column_proportion)
        self.treeView.setColumnWidth(0, first_column_width)
        if col_count > 1:
            remaining = max(total_width - first_column_width, 0)
            per_col = remaining // (col_count - 1) if col_count > 1 else 0
            for column in range(1, col_count):
                self.treeView.setColumnWidth(column, per_col)
        self._save_current_column_widths()

    def resizeEvent(self, event: QResizeEvent):
        """
        Resize event

        :param event: Event object
        """
        super().resizeEvent(event)
        if event.oldSize().width() != event.size().width():
            self.adjustColumnWidths()

    def openContextMenu(self, position):
        """
        Open context menu

        :param position: mouse position
        """
        paths = self._selected_paths()
        if paths:
            first_path = paths[0]
            multiple = len(paths) > 1
            target_multi = paths if multiple else first_path
            actions = {}
            preview_actions = []
            use_actions = []

            can_preview = False
            try:
                can_preview = self.window.core.filesystem.actions.has_preview(target_multi)
            except Exception:
                try:
                    can_preview = self.window.core.filesystem.actions.has_preview(first_path)
                except Exception:
                    can_preview = False

            if can_preview:
                try:
                    preview_actions = self.window.core.filesystem.actions.get_preview(self, target_multi)
                except Exception:
                    try:
                        preview_actions = self.window.core.filesystem.actions.get_preview(self, first_path)
                    except Exception:
                        preview_actions = []

            parent = self._parent_for_selection(paths)

            actions['open'] = QAction(self._icons['open'], trans('action.open'), self)
            actions['open'].triggered.connect(lambda: self.action_open(target_multi))

            actions['open_dir'] = QAction(self._icons['open_dir'], trans('action.open_dir'), self)
            actions['open_dir'].triggered.connect(lambda: self.action_open_dir(target_multi))

            actions['download'] = QAction(self._icons['download'], trans('action.download'), self)
            actions['download'].triggered.connect(lambda: self.window.controller.files.download_local(target_multi))

            actions['rename'] = QAction(self._icons['rename'], trans('action.rename'), self)
            actions['rename'].triggered.connect(lambda: self.action_rename(target_multi))

            actions['duplicate'] = QAction(self._icons['duplicate'], trans('action.duplicate'), self)
            actions['duplicate'].triggered.connect(lambda: self.window.controller.files.duplicate_local(target_multi, ""))

            actions['touch'] = QAction(self._icons['touch'], trans('action.touch'), self)
            actions['touch'].triggered.connect(lambda: self.window.controller.files.touch_file(parent))

            actions['mkdir'] = QAction(self._icons['mkdir'], trans('action.mkdir'), self)
            actions['mkdir'].triggered.connect(lambda: self.action_make_dir(parent))

            actions['refresh'] = QAction(self._icons['refresh'], trans('action.refresh'), self)
            actions['refresh'].triggered.connect(lambda: self.window.controller.files.update_explorer())

            actions['upload'] = QAction(self._icons['upload'], trans('action.upload'), self)
            actions['upload'].triggered.connect(lambda: self.window.controller.files.upload_local(parent))

            actions['delete'] = QAction(self._icons['delete'], trans('action.delete'), self)
            actions['delete'].triggered.connect(lambda: self.action_delete(target_multi))

            actions['copy'] = QAction(self._icons['copy'], trans('action.copy'), self)
            actions['copy'].triggered.connect(self.action_copy_selection)
            actions['cut'] = QAction(self._icons['cut'], trans('action.cut'), self)
            actions['cut'].triggered.connect(self.action_cut_selection)
            actions['paste'] = QAction(self._icons['paste'], trans('action.paste'), self)
            actions['paste'].triggered.connect(lambda: self.action_paste_into(parent))
            actions['paste'].setEnabled(self._can_paste())

            # Pack / Unpack availability
            try:
                can_unpack_all = all(
                    os.path.isfile(p) and self.window.core.filesystem.packer.can_unpack(p)
                    for p in paths
                )
            except Exception:
                can_unpack_all = False

            # Build menu
            menu = QMenu(self)
            if preview_actions:
                for action in preview_actions:
                    menu.addAction(action)
            menu.addAction(actions['open'])
            menu.addAction(actions['open_dir'])

            use_menu = QMenu(trans('action.use'), self)

            files_only = all(os.path.isfile(p) for p in paths)
            if files_only:
                actions['use_attachment'] = QAction(self._icons['attachment'], trans('action.use.attachment'), self)
                actions['use_attachment'].triggered.connect(
                    lambda: self.window.controller.files.use_attachment(target_multi)
                )
                use_menu.addAction(actions['use_attachment'])

            if self.window.core.filesystem.actions.has_use(first_path):
                use_actions = self.window.core.filesystem.actions.get_use(self, first_path)

            if use_actions:
                for action in use_actions:
                    use_menu.addAction(action)

            actions['use_copy_work_path'] = QAction(self._icons['copy'], trans('action.use.copy_work_path'), self)
            actions['use_copy_work_path'].triggered.connect(
                lambda: self.window.controller.files.copy_work_path(target_multi)
            )

            actions['use_copy_sys_path'] = QAction(self._icons['copy'], trans('action.use.copy_sys_path'), self)
            actions['use_copy_sys_path'].triggered.connect(
                lambda: self.window.controller.files.copy_sys_path(target_multi)
            )

            actions['use_read_cmd'] = QAction(self._icons['read'], trans('action.use.read_cmd'), self)
            actions['use_read_cmd'].triggered.connect(
                lambda: self.window.controller.files.make_read_cmd(target_multi)
            )

            use_menu.addAction(actions['use_copy_work_path'])
            use_menu.addAction(actions['use_copy_sys_path'])
            use_menu.addAction(actions['use_read_cmd'])
            menu.addMenu(use_menu)

            allowed_any = any(self.window.core.idx.indexing.is_allowed(p) for p in paths)
            if allowed_any:
                idx_menu = QMenu(trans('action.idx'), self)
                idx_list = self.window.core.config.get('llama.idx.list')
                if len(idx_list) > 0:
                    for idx in idx_list:
                        id = idx['id']
                        name = f"{idx['name']} ({idx['id']})"
                        action = QAction(self._icons['db'], f"IDX: {name}", self)
                        action.triggered.connect(lambda checked=False, id=id, target=target_multi: self.action_idx(target, id))
                        idx_menu.addAction(action)

                remove_idx_set = set()
                for p in paths:
                    status = self.model.get_index_status(p)
                    if status.get('indexed'):
                        for ix in status.get('indexed_in', []):
                            remove_idx_set.add(ix)

                if len(remove_idx_set) > 0:
                    idx_menu.addSeparator()
                    for ix in sorted(remove_idx_set):
                        action = QAction(self._icons['delete'], trans("action.idx.remove") + ": " + ix, self)
                        action.triggered.connect(
                            lambda checked=False, ix=ix, target=target_multi: self.action_idx_remove(target, ix)
                        )
                        idx_menu.addAction(action)

                menu.addMenu(idx_menu)

            menu.addSeparator()
            menu.addAction(actions['copy'])
            menu.addAction(actions['cut'])
            menu.addAction(actions['paste'])
            menu.addSeparator()

            # Pack submenu (available for any selection)
            pack_menu = QMenu(trans("action.pack"), self)
            a_zip = QAction(self._icons['pack'], "ZIP (.zip)", self)
            a_zip.triggered.connect(lambda: self.action_pack(target_multi, 'zip'))
            a_tgz = QAction(self._icons['pack'], "Tar GZip (.tar.gz)", self)
            a_tgz.triggered.connect(lambda: self.action_pack(target_multi, 'tar.gz'))
            pack_menu.addAction(a_zip)
            pack_menu.addAction(a_tgz)
            menu.addMenu(pack_menu)

            # Unpack (only when all selected are supported archives)
            if can_unpack_all:
                a_unpack = QAction(self._icons['unpack'], trans("action.unpack"), self)
                a_unpack.triggered.connect(lambda: self.action_unpack(target_multi))
                menu.addAction(a_unpack)

            menu.addSeparator()
            menu.addAction(actions['download'])
            menu.addAction(actions['touch'])
            menu.addAction(actions['mkdir'])
            menu.addAction(actions['upload'])
            menu.addAction(actions['refresh'])

            menu.addAction(actions['rename'])
            menu.addAction(actions['duplicate'])
            menu.addAction(actions['delete'])

            menu.exec(QCursor.pos())
        else:
            actions = {}

            actions['touch'] = QAction(self._icons['touch'], trans('action.touch'), self)
            actions['touch'].triggered.connect(
                lambda: self.window.controller.files.touch_file(self.directory),
            )

            actions['open_dir'] = QAction(self._icons['open_dir'], trans('action.open_dir'), self)
            actions['open_dir'].triggered.connect(
                lambda: self.action_open_dir(self.directory),
            )

            actions['mkdir'] = QAction(self._icons['mkdir'], trans('action.mkdir'), self)
            actions['mkdir'].triggered.connect(
                lambda: self.action_make_dir(self.directory),
            )

            actions['upload'] = QAction(self._icons['upload'], trans('action.upload'), self)
            actions['upload'].triggered.connect(
                lambda: self.window.controller.files.upload_local(),
            )

            actions['paste'] = QAction(self._icons['paste'], trans('action.paste'), self)
            actions['paste'].triggered.connect(lambda: self.action_paste_into(self.directory))
            actions['paste'].setEnabled(self._can_paste())

            menu = QMenu(self)
            menu.addAction(actions['touch'])
            menu.addAction(actions['open_dir'])
            menu.addAction(actions['mkdir'])
            menu.addAction(actions['upload'])
            menu.addAction(actions['paste'])
            menu.exec(QCursor.pos())

    def action_open(self, path: Union[str, list]):
        """
        Open action handler

        :param path: path to open (str or list of str)
        """
        self.window.controller.files.open(path)

    def action_idx(self, path: Union[str, list], idx: str):
        """
        Index file or dir handler

        :param path: path to open (str or list of str)
        :param idx: index ID to use (name)
        """
        self.window.controller.idx.indexer.index_file(path, idx)

    def action_idx_remove(self, path: Union[str, list], idx: str):
        """
        Remove file or dir from index handler

        :param path: path to open (str or list of str)
        :param idx: index ID to use (name)
        """
        self.window.controller.idx.indexer.index_file_remove(path, idx)

    def action_open_dir(self, path: Union[str, list]):
        """
        Open in directory action handler

        :param path: path to open (str or list of str)
        """
        self.window.controller.files.open_dir(path, True)

    def action_make_dir(self, path: str = None):
        """
        Make directory action handler

        :param path: parent path
        """
        self.window.controller.files.make_dir_dialog(path)

    def action_rename(self, path: Union[str, list]):
        """
        Rename action handler

        :param path: path to rename (str or list of str)
        """
        self.window.controller.files.rename(path)

    def action_delete(self, path: Union[str, list]):
        """
        Delete action handler

        :param path: path to delete (str or list of str)
        """
        self.window.controller.files.delete(path)

    def action_pack(self, path: Union[str, list], fmt: str):
        """
        Pack selected items into an archive.

        :param path: path or list of paths to include
        :param fmt: 'zip' or 'tar.gz'
        """
        paths = path if isinstance(path, list) else [path]
        try:
            dst = self.window.core.filesystem.packer.pack_paths(paths, fmt)
        except Exception as e:
            try:
                self.window.core.debug.log(e)
            except Exception:
                pass
            dst = None

        try:
            self.window.controller.files.update_explorer()
        except Exception:
            self.update_view()

        if dst and os.path.exists(dst):
            self._reveal_paths([dst], select_first=True)

    def action_unpack(self, path: Union[str, list]):
        """
        Unpack selected archives to sibling directories named after archives.

        :param path: path or list of paths to archives
        """
        paths = path if isinstance(path, list) else [path]
        created = []
        for p in paths:
            try:
                if self.window.core.filesystem.packer.can_unpack(p):
                    out_dir = self.window.core.filesystem.packer.unpack_to_sibling_dir(p)
                    if out_dir:
                        created.append(out_dir)
            except Exception as e:
                try:
                    self.window.core.debug.log(e)
                except Exception:
                    pass

        try:
            self.window.controller.files.update_explorer()
        except Exception:
            self.update_view()

        if created:
            self._reveal_paths(created, select_first=True)

    # ===== Copy / Cut / Paste API =====

    def _selected_paths(self) -> list:
        """Return unique selected file system paths from first column."""
        paths = []
        try:
            indexes = self.treeView.selectionModel().selectedRows(0)
        except Exception:
            indexes = []
        for idx in indexes:
            try:
                p = self.model.filePath(idx)
                if p and p not in paths:
                    paths.append(p)
            except Exception:
                continue
        return paths

    def _parent_for_selection(self, paths: list) -> str:
        """
        Determine a sensible parent directory for operations like paste/touch/mkdir when selection may contain many items.
        - For single selection: item if it is a directory; otherwise its parent directory.
        - For multi selection: common parent directory of all selected items; falls back to explorer root if not determinable.
        """
        if not paths:
            return self.directory
        if len(paths) == 1:
            p = paths[0]
            return p if os.path.isdir(p) else os.path.dirname(p)
        try:
            parents = [p if os.path.isdir(p) else os.path.dirname(p) for p in paths]
            cp = os.path.commonpath([os.path.abspath(x) for x in parents])
            return cp if os.path.isdir(cp) else os.path.dirname(cp)
        except Exception:
            return self.directory

    def action_copy_selection(self):
        """Copy currently selected files/dirs to system clipboard and internal buffer."""
        paths = self._selected_paths()
        if not paths:
            return
        self._set_clipboard_files(paths, mode='copy')

    def action_cut_selection(self):
        """Cut currently selected files/dirs to system clipboard and internal buffer (virtual until paste)."""
        paths = self._selected_paths()
        if not paths:
            return
        self._set_clipboard_files(paths, mode='cut')

    def action_paste_into(self, target_dir: str):
        """Paste clipboard files into given directory and expand/scroll to them."""
        if not target_dir:
            target_dir = self.directory
        paths, mode = self._get_clipboard_files_and_mode()
        if not paths:
            return

        dest_paths = []
        try:
            if mode == 'cut':
                dest_paths = self._move_paths(paths, target_dir)
            else:
                dest_paths = self._copy_paths(paths, target_dir)
        finally:
            self._cb_paths = []
            self._cb_mode = None

        if os.path.isdir(target_dir):
            self._expand_dir(target_dir, center=False)

        try:
            self.window.controller.files.update_explorer()
        except Exception:
            self.update_view()

        if dest_paths:
            self._reveal_paths(dest_paths, select_first=True)

    def action_paste_into_current(self):
        """Paste into directory derived from current selection or root when none."""
        try:
            sel = self.treeView.selectionModel()
            indexes = sel.selectedRows(0) if sel is not None else []
        except Exception:
            indexes = []
        if indexes:
            path = self.model.filePath(indexes[0])
            target = path if os.path.isdir(path) else os.path.dirname(path)
        else:
            target = self.directory
        self.action_paste_into(target)

    def _can_paste(self) -> bool:
        """Check if there are files to paste either from system clipboard or internal."""
        try:
            md = QGuiApplication.clipboard().mimeData()
            if md and md.hasUrls():
                for url in md.urls():
                    if url.isLocalFile():
                        return True
        except Exception:
            pass
        return bool(self._cb_paths)

    # ===== Filesystem helpers =====

    def _copy_paths(self, paths: list, target_dir: str):
        """Copy each path into target_dir, directories are copied recursively. Returns list of new destinations."""
        dests = []
        if not os.path.isdir(target_dir):
            os.makedirs(target_dir, exist_ok=True)
        for src in paths:
            try:
                if not os.path.exists(src):
                    continue
                base_name = os.path.basename(src.rstrip(os.sep))
                dst = self._unique_dest(target_dir, base_name)
                if os.path.isdir(src):
                    shutil.copytree(src, dst, copy_function=shutil.copy2)
                else:
                    shutil.copy2(src, dst)
                dests.append(dst)
            except Exception as e:
                try:
                    self.window.core.debug.log(e)
                except Exception:
                    pass
        return dests

    def _move_paths(self, paths: list, target_dir: str):
        """Move each path into target_dir. Skips invalid moves (into itself). Returns list of new destinations."""
        dests = []
        if not os.path.isdir(target_dir):
            os.makedirs(target_dir, exist_ok=True)
        for src in paths:
            try:
                if not os.path.exists(src):
                    continue
                if os.path.isdir(src):
                    try:
                        sp = os.path.abspath(src)
                        tp = os.path.abspath(target_dir)
                        if os.path.commonpath([sp]) == os.path.commonpath([sp, tp]):
                            continue
                    except Exception:
                        pass
                base_name = os.path.basename(src.rstrip(os.sep))
                dst = os.path.join(target_dir, base_name)
                if os.path.abspath(dst) == os.path.abspath(src):
                    continue
                if os.path.exists(dst):
                    dst = self._unique_dest(target_dir, base_name)
                shutil.move(src, dst)
                dests.append(dst)
            except Exception as e:
                try:
                    self.window.core.debug.log(e)
                except Exception:
                    pass
        return dests

    def _unique_dest(self, target_dir: str, name: str) -> str:
        """Return a unique destination path in target_dir based on name."""
        root, ext = os.path.splitext(name)
        candidate = os.path.join(target_dir, name)
        if not os.path.exists(candidate):
            return candidate
        i = 1
        while True:
            suffix = " - Copy" if i == 1 else f" - Copy ({i})"
            cand = os.path.join(target_dir, f"{root}{suffix}{ext}")
            if not os.path.exists(cand):
                return cand
            i += 1

    def _expand_dir(self, path: str, center: bool = False):
        """Expand and optionally center on a directory index."""
        try:
            idx = self.model.index(path)
            if idx.isValid():
                if not self.treeView.isExpanded(idx):
                    self.treeView.expand(idx)
                if center:
                    self.treeView.scrollTo(idx, QTreeView.PositionAtCenter)
                else:
                    self.treeView.scrollTo(idx, QTreeView.EnsureVisible)
        except Exception:
            pass

    def _reveal_paths(self, paths: list, select_first: bool = True):
        """
        Reveal and optionally select the given paths in the view.
        Tries a few times with small delays to wait for model refresh.
        """
        def do_reveal(attempts_left=6):
            try:
                sm = self.treeView.selectionModel()
            except Exception:
                sm = None
            first_index = None
            for p in paths:
                dir_path = p if os.path.isdir(p) else os.path.dirname(p)
                self._expand_dir(dir_path, center=False)
                idx = self.model.index(p)
                if idx.isValid():
                    if first_index is None:
                        first_index = idx
                    self.treeView.scrollTo(idx, QTreeView.PositionAtCenter)
            if first_index is not None and sm is not None and select_first:
                try:
                    sm.clearSelection()
                    self.treeView.setCurrentIndex(first_index)
                    self.treeView.scrollTo(first_index, QTreeView.PositionAtCenter)
                except Exception:
                    pass
            elif attempts_left > 0:
                QTimer.singleShot(150, lambda: do_reveal(attempts_left - 1))

        QTimer.singleShot(100, do_reveal)

    # ===== Clipboard integration (OS + internal) =====

    def _urls_to_text_uri_list(self, urls):
        """
        Build RFC compliant text/uri-list payload (CRLF separated).
        """
        parts = []
        for u in urls:
            try:
                parts.append(u.toString(QUrl.FullyEncoded))
            except Exception:
                parts.append(u.toString())
        data = ("\r\n".join(parts) + "\r\n").encode("utf-8")
        return data

    def _build_gnome_payload(self, urls, verb: str):
        """
        Build x-special/gnome-copied-files payload:
        copy|cut + newline + list of file:// URLs + trailing newline.
        """
        lines = [verb]
        for u in urls:
            try:
                lines.append(u.toString(QUrl.FullyEncoded))
            except Exception:
                lines.append(u.toString())
        return ("\n".join(lines) + "\n").encode("utf-8")

    def _set_clipboard_files(self, paths: list, mode: str = 'copy'):
        """
        Set system clipboard with file urls and cut/copy semantics; keep internal buffer.
        Designed to work across Linux (GNOME/KDE), Windows, and macOS as far as OS allows.
        """
        self._cb_paths = [os.path.abspath(p) for p in paths if p]
        self._cb_mode = 'cut' if mode == 'cut' else 'copy'

        try:
            urls = [QUrl.fromLocalFile(p) for p in self._cb_paths]
            md = QMimeData()

            md.setData("text/uri-list", self._urls_to_text_uri_list(urls))
            md.setUrls(urls)

            try:
                md.setData("application/x-kde-cutselection", b"1" if self._cb_mode == 'cut' else b"0")
            except Exception:
                pass

            try:
                verb = "cut" if self._cb_mode == 'cut' else "copy"
                payload = self._build_gnome_payload(urls, verb)
                md.setData("x-special/gnome-copied-files", payload)
                md.setData("x-special/nautilus-clipboard", payload)
            except Exception:
                pass

            try:
                effect = 2 if self._cb_mode == 'cut' else 1
                data = struct.pack("<I", effect)
                md.setData('application/x-qt-windows-mime;value="Preferred DropEffect"', data)
                md.setData("application/x-qt-windows-mime;value=Preferred DropEffect", data)
                md.setData("Preferred DropEffect", data)
            except Exception:
                pass

            cb = QGuiApplication.clipboard()
            cb.setMimeData(md, QClipboard.Clipboard)
            try:
                cb.setMimeData(md, QClipboard.Selection)
            except Exception:
                pass
        except Exception as e:
            try:
                self.window.core.debug.log(e)
            except Exception:
                pass

    def _get_clipboard_files_and_mode(self):
        """
        Read file urls and cut/copy mode from system clipboard.
        Returns tuple (paths, mode) where mode in {'copy','cut'}.
        Falls back to internal buffer if system clipboard does not provide file urls.
        """
        paths = []
        mode = 'copy'
        try:
            md = QGuiApplication.clipboard().mimeData()
        except Exception:
            md = None

        try:
            if md:
                urls = []
                if md.hasUrls():
                    urls = md.urls()
                elif md.hasFormat("text/uri-list"):
                    try:
                        raw = bytes(md.data("text/uri-list")).decode("utf-8", "ignore")
                        for line in raw.splitlines():
                            line = line.strip()
                            if line and not line.startswith("#"):
                                u = QUrl(line)
                                if u.isLocalFile():
                                    urls.append(u)
                    except Exception:
                        pass

                for u in urls:
                    try:
                        if u.isLocalFile():
                            lf = u.toLocalFile()
                            if lf:
                                paths.append(lf)
                    except Exception:
                        continue

                try:
                    if md.hasFormat("application/x-kde-cutselection"):
                        data = bytes(md.data("application/x-kde-cutselection"))
                        if data and (data.startswith(b'1') or data == b"\x01"):
                            mode = 'cut'
                except Exception:
                    pass
                try:
                    if md.hasFormat("x-special/gnome-copied-files"):
                        data = bytes(md.data("x-special/gnome-copied-files")).decode("utf-8", "ignore")
                        if data.splitlines()[0].strip().lower().startswith("cut"):
                            mode = 'cut'
                    elif md.hasFormat("x-special/nautilus-clipboard"):
                        data = bytes(md.data("x-special/nautilus-clipboard")).decode("utf-8", "ignore")
                        if data.splitlines()[0].strip().lower().startswith("cut"):
                            mode = 'cut'
                except Exception:
                    pass
                try:
                    for key in ('application/x-qt-windows-mime;value="Preferred DropEffect"',
                                "application/x-qt-windows-mime;value=Preferred DropEffect",
                                "Preferred DropEffect"):
                        if md.hasFormat(key):
                            data = bytes(md.data(key))
                            if data and len(data) >= 4:
                                value = struct.unpack("<I", data[:4])[0]
                                if value & 2:
                                    mode = 'cut'
                                    break
                except Exception:
                    pass
        except Exception:
            paths = []

        if not paths and self._cb_paths:
            paths = list(self._cb_paths)
            mode = self._cb_mode or 'copy'

        return paths, mode


class IndexedFileSystemModel(QFileSystemModel):
    def __init__(self, window, index_dict, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.window = window
        self.index_dict = index_dict
        self._status_cache = {}
        self.directoryLoaded.connect(self.refresh_path)
        try:
            self.setReadOnly(False)
        except Exception:
            pass

    def refresh_path(self, path):
        index = self.index(path)
        if index.isValid():
            self._status_cache.clear()
            self.dataChanged.emit(index, index)

    def columnCount(self, parent=QModelIndex()) -> int:
        """
        Return column count

        :param parent: parent
        :return: column count
        """
        return super().columnCount(parent) + 1

    def data(self, index, role=Qt.DisplayRole) -> any:
        """
        Data handler

        :param index: row index
        :param role: role
        :return: data
        """
        last_col = self.columnCount() - 1
        if index.column() == last_col:
            if role == Qt.DisplayRole:
                file_path = self.filePath(index.siblingAtColumn(0))
                status = self.get_index_status(file_path)
                if status['indexed']:
                    ts = status['last_index_at']
                    dt = datetime.datetime.fromtimestamp(ts)
                    if dt.date() == datetime.date.today():
                        ds = dt.strftime("%H:%M")
                    else:
                        ds = dt.strftime("%Y-%m-%d %H:%M")
                    content = f"{ds} ({','.join(status['indexed_in'])})"
                else:
                    content = '-'
                return content
        elif index.column() == last_col - 1:
            if role == Qt.DisplayRole:
                dt_qt = self.lastModified(index)
                ts = dt_qt.toSecsSinceEpoch()
                dt_py = datetime.datetime.fromtimestamp(ts)
                if dt_py.date() == datetime.date.today():
                    data = dt_py.strftime("%H:%M")
                else:
                    data = dt_py.strftime("%Y-%m-%d %H:%M")
                file_path = self.filePath(index.siblingAtColumn(0))
                status = self.get_index_status(file_path)
                if status['indexed']:
                    if 'last_index_at' in status and status['last_index_at'] < ts:
                        data += '*'
                return data

        return super().data(index, role)

    def get_index_status(self, file_path) -> dict:
        """
        Get index status

        :param file_path: file path
        :return: file index status
        """
        file_id = self.window.core.idx.files.get_id(file_path)
        cached = self._status_cache.get(file_id)
        if cached is not None:
            return cached
        indexed_in = []
        indexed_timestamps = {}
        last_index_at = 0
        for idx in self.index_dict:
            items = self.index_dict[idx]
            if file_id in items:
                indexed_in.append(idx)
                ts = items[file_id]['indexed_ts']
                indexed_timestamps[idx] = ts
                if ts > last_index_at:
                    last_index_at = ts
        if indexed_in:
            indexed_in.sort(key=lambda x: indexed_timestamps[x], reverse=True)
            result = {
                'indexed': True,
                'indexed_in': indexed_in,
                'last_index_at': last_index_at,
            }
        else:
            result = {'indexed': False}
        self._status_cache[file_id] = result
        return result

    def headerData(self, section, orientation, role=Qt.DisplayRole) -> str:
        """
        Prepare Header data (append Indexed column)

        :param section: Section
        :param orientation: Orientation
        :param role: Role
        :return: Header data
        """
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            if section == 0:  # name
                return trans('files.explorer.header.name')
            elif section == 1:  # size
                return trans('files.explorer.header.size')
            elif section == 2:  # type
                return trans('files.explorer.header.type')
            elif section == 3:  # modified
                return trans('files.explorer.header.modified')
            elif section == 4:  # indexed
                return trans('files.explorer.header.indexed')
        return super().headerData(section, orientation, role)

    def update_idx_status(self, idx_data):
        """
        Update index data status

        :param idx_data: new index data dict
        """
        self.index_dict = idx_data
        self._status_cache.clear()
        row_count = self.rowCount()
        if row_count > 0:
            top_left_index = self.index(0, 0)
            bottom_right_index = self.index(row_count - 1, self.columnCount() - 1)
            self.dataChanged.emit(top_left_index, bottom_right_index, [Qt.DisplayRole])
        path = self.rootPath()
        self.setRootPath("")
        self.setRootPath(path)
        self.layoutChanged.emit()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.11 18:00:00                  #
# ================================================== #

from PySide6 import QtCore
from PySide6.QtWidgets import QLineEdit

class ConsoleInput(QLineEdit):
    def __init__(self, window=None):
        """
        Console input

        :param window: main window
        """
        super(ConsoleInput, self).__init__(window)
        self.window = window
        self.setPlaceholderText("Console... Type your command here")
        self.setProperty('class', 'text-editor')
        self.setFocus()

    def keyPressEvent(self, event):
        """
        Key press event

        :param event: key event
        """
        handled = False
        if event.key() in (QtCore.Qt.Key_Return, QtCore.Qt.Key_Enter):
            self.window.core.debug.console.on_send()
            handled = True
            self.setFocus()
        if not handled:
            super(ConsoleInput, self).keyPressEvent(event)



# ui/layout/presets.py
#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.26 13:30:00                  #
# ================================================== #

from PySide6 import QtCore
from PySide6.QtGui import QStandardItemModel, QStandardItem, Qt, QIcon
from PySide6.QtWidgets import QVBoxLayout, QHBoxLayout, QPushButton, QWidget, QSizePolicy

from pygpt_net.core.types import (
    MODE_AGENT,
    MODE_EXPERT,
)
from pygpt_net.ui.widget.element.labels import HelpLabel, TitleLabel
from pygpt_net.ui.widget.lists.preset import PresetList

from pygpt_net.ui.layout.toolbox.footer import Footer
from pygpt_net.utils import trans

class Presets:
    def __init__(self, window=None):
        """
        Toolbox UI

        :param window: Window instance
        """
        self.window = window
        self.footer = Footer(window)
        self.id = 'preset.presets'

    def setup(self) -> QWidget:
        """
        Setup presets

        :return: QWidget
        """
        presets = self.setup_presets()

        self.window.ui.nodes['presets.widget'] = QWidget()
        self.window.ui.nodes['presets.widget'].setLayout(presets)
        self.window.ui.nodes['presets.widget'].setMinimumHeight(180)
        self.window.ui.nodes['presets.widget'].setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        return self.window.ui.nodes['presets.widget']

    def setup_presets(self) -> QVBoxLayout:
        """
        Setup list

        :return: QVBoxLayout
        """
        nodes = self.window.ui.nodes

        nodes['preset.presets.new'] = QPushButton(QIcon(":/icons/add.svg"), "")
        nodes['preset.presets.new'].clicked.connect(
            lambda _=False: self.window.controller.presets.editor.edit()
        )

        nodes['preset.presets.label'] = TitleLabel(trans("toolbox.presets.label"))
        nodes['preset.agents.label'] = TitleLabel(trans("toolbox.agents.label"))
        nodes['preset.experts.label'] = TitleLabel(trans("toolbox.experts.label"))
        nodes['preset.presets.label'].setVisible(False)
        nodes['preset.agents.label'].setVisible(False)
        nodes['preset.experts.label'].setVisible(False)

        header = QHBoxLayout()
        header.addWidget(nodes['preset.presets.label'])
        header.addWidget(nodes['preset.agents.label'])
        header.addWidget(nodes['preset.experts.label'])
        header.addWidget(nodes['preset.presets.new'], alignment=Qt.AlignRight)
        header.setContentsMargins(5, 0, 0, 0)

        nodes[self.id] = PresetList(self.window, self.id)
        nodes[self.id].selection_locked = self.window.controller.presets.preset_change_locked
        nodes[self.id].setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        nodes['tip.toolbox.presets'] = HelpLabel(trans('tip.toolbox.presets'), self.window)
        nodes['tip.toolbox.presets'].setAlignment(Qt.AlignCenter)

        layout = QVBoxLayout()
        layout.addStretch()
        layout.addLayout(header)
        layout.addWidget(nodes[self.id], 1)
        layout.addWidget(nodes['tip.toolbox.presets'])
        layout.setContentsMargins(2, 5, 5, 5)

        self.window.ui.models[self.id] = self.create_model(self.window)
        nodes[self.id].setModel(self.window.ui.models[self.id])

        return layout

    def create_model(self, parent) -> QStandardItemModel:
        """
        Create list model
        :param parent: parent widget
        :return: QStandardItemModel
        """
        return QStandardItemModel(0, 1, parent)

    def update_presets(self, data):
        """
        Update presets list

        :param data: Data to update
        """
        mode = self.window.core.config.get('mode')
        nodes = self.window.ui.nodes
        models = self.window.ui.models

        view: PresetList = nodes[self.id]
        model = models.get(self.id)

        # If view requested selection override, do NOT override it by backup
        selection_override_ids = getattr(view, "_selection_override_ids", None)
        if not selection_override_ids:
            view.backup_selection()

        if model is None:
            model = self.create_model(self.window)
            models[self.id] = model
            view.setModel(model)

        # Preserve current scroll position across model rebuild to avoid a visible jump to the top.
        # This is applied while updates are disabled, then restored just before re-enabling them.
        try:
            v = view.verticalScrollBar().value()
            h = view.horizontalScrollBar().value()
            view.set_pending_v_scroll(v)
            view.set_pending_h_scroll(h)
        except Exception:
            pass

        # Block user input during model rebuild to avoid crashes on quick clicks
        view.begin_model_update()

        # Turn off updates to avoid flicker and transient artifacts
        view.setUpdatesEnabled(False)
        try:
            # Rebuild model cleanly to avoid any stale items causing visual glitches
            model.clear()
            model.setColumnCount(1)

            if data:
                is_expert_mode = (mode == MODE_EXPERT)
                is_agent_mode = (mode == MODE_AGENT)
                count_experts = self.window.core.experts.count_experts if is_agent_mode else None
                startswith_current = "current."

                role_uuid = QtCore.Qt.UserRole + 1
                role_id = QtCore.Qt.UserRole + 2
                role_is_special = QtCore.Qt.UserRole + 3

                for i, (key, item) in enumerate(data.items()):
                    qitem = QStandardItem()

                    name = item.name
                    if is_expert_mode and item.enabled and not key.startswith(startswith_current):
                        name = f"[x] {name}"
                    elif is_agent_mode:
                        num_experts = count_experts(key)
                        if num_experts > 0:
                            name = f"{name} ({num_experts} experts)"

                    prompt = str(item.prompt)
                    tooltip = prompt if len(prompt) <= 80 else f"{prompt[:80]}..."

                    qitem.setText(name)
                    qitem.setToolTip(tooltip)
                    qitem.setData(item.uuid, role_uuid)
                    qitem.setData(key, role_id)
                    qitem.setData(key.startswith(startswith_current), role_is_special)

                    # Pin row 0 (no drag, no drop)
                    # Other rows: drag enabled only; drop is handled by view between rows
                    if i != 0 and not key.startswith(startswith_current):
                        flags = Qt.ItemIsEnabled | Qt.ItemIsSelectable | Qt.ItemIsDragEnabled
                    else:
                        flags = Qt.ItemIsEnabled | Qt.ItemIsSelectable
                    qitem.setFlags(flags)

                    model.appendRow(qitem)
        finally:
            # Apply pending scroll (if any) before re-enabling updates
            view.apply_pending_scroll()
            view.setUpdatesEnabled(True)

        dnd_enabled = bool(self.window.core.config.get('presets.drag_and_drop.enabled'))
        view.set_dnd_enabled(dnd_enabled)

        # If override requested, force saved selection IDs to those override IDs
        if selection_override_ids:
            view._saved_selection_ids = list(selection_override_ids)
            view._selection_override_ids = None  # consume one-shot override

        # Restore selection by ID (so it follows the same item even if rows moved)
        view.restore_selection()
        # Force repaint in case Qt defers layout until next input
        view.viewport().update()

        # Clear one-shot pending scroll values and re-enable user interaction
        view.clear_pending_scroll()
        view.end_model_update()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.16 22:00:00                  #
# ================================================== #

from typing import Any

from PySide6.QtCore import QEvent
from PySide6.QtWidgets import QTabWidget, QWidget

from pygpt_net.core.tabs.tab import Tab


class TabBody(QTabWidget):
    def __init__(self, window=None):
        super(TabBody, self).__init__()
        self.window = window
        self.owner = None
        self.body = []
        self.refs = []
        self.on_delete = None  # callback on delete
        self.installEventFilter(self)

    def cleanup(self):
        """
        Clean up on delete
        """
        if self.on_delete and callable(self.on_delete):
            self.on_delete(self)
        self.delete_refs()

    def unwrap(self, widget: QWidget) -> None:
        """
        Remove widget from tab body

        :param widget: widget to remove
        """
        layout = self.layout()
        if layout is not None:
            layout.removeWidget(widget)
        self.delete_ref(widget)

    def unwrap_all(self) -> None:
        """
        Remove all widgets from tab body
        """
        layout = self.layout()
        if layout is not None:
            while layout.count():
                item = layout.takeAt(0)
                widget = item.widget()
                try:
                    layout.removeWidget(widget)
                except Exception:
                    pass
                try:
                    self.delete_ref(widget)
                except Exception:
                    pass

    def add_ref(self, ref: Any) -> None:
        """
        Add reference to widget in this tab

        :param ref: widget reference
        """
        if ref not in self.refs:
            self.refs.append(ref)

    def delete_refs(self) -> None:
        """
        Delete all references to widgets in this tab
        """
        for ref in self.refs:
            if ref is None:
                continue
            if ref and hasattr(ref, 'on_delete'):
                try:
                    ref.on_delete()
                except Exception:
                    pass
            if ref and hasattr(ref, 'deleteLater'):
                try:
                    ref.deleteLater()
                except Exception:
                    pass
        del self.refs[:]

    def delete_ref(self, widget: Any) -> None:
        """
        Unpin reference to widget in this tab

        :param widget: widget reference
        """
        for ref in self.refs:
            if ref and ref is widget:
                self.refs.remove(ref)
                break

    def append(self, body: QWidget):
        """
        Append tab body (parent widget)

        :param body: tab body widget
        """
        self.body.append(body)

    def get_body(self):
        """
        Get tab body (parent widget)

        :return: tab body widget
        """
        return self.body

    def attach_tab(self, tab: Tab):
        """
        Attach tab to body

        :param tab: tab instance
        """
        for body in self.body:
            if hasattr(body, 'set_tab'):
                body.set_tab(tab)

    def setOwner(self, owner: Tab):
        """
        Set tab parent (owner)

        :param owner: parent tab instance
        """
        self.owner = owner
        self.attach_tab(owner)

    def getOwner(self) -> Tab:
        """
        Get tab parent (owner)

        :return: parent tab instance
        """
        return self.owner

    def eventFilter(self, source, event):
        """
        Focus event filter

        :param source: source
        :param event: event
        """
        t = event.type()
        if (t == QEvent.ChildAdded and
                source is self and
                event.child().isWidgetType()):
            self._glwidget = event.child()
            self._glwidget.installEventFilter(self)
        elif t == event.Type.MouseButtonPress:
            # handle column focus
            if self.owner is not None:
                col_idx = self.owner.column_idx
                self.window.controller.ui.tabs.on_column_focus(col_idx)
        return super().eventFilter(source, event)

    def to_dict(self) -> dict:
        """
        Convert to dict

        :return: dict
        """
        return {
            "refs": [str(ref) for ref in self.refs],  # references to widgets
            "body": [str(b) for b in self.body],  # body widgets
            "len(layout)": self.layout().count() if self.layout() else 0,
        }


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2026.01.04 19:00:00                  #
# ================================================== #

from __future__ import annotations

import json
from typing import Optional, Dict, Any, List, Tuple

from pygpt_net.core.types import MODE_CHAT
from pygpt_net.core.bridge.context import BridgeContext, MultimodalContext
from pygpt_net.item.attachment import AttachmentItem
from pygpt_net.item.ctx import CtxItem
from pygpt_net.item.model import ModelItem

# xAI SDK chat helpers (system/user/assistant/image) for message building
from xai_sdk.chat import (
    system as xsystem,
    user as xuser,
    assistant as xassistant,
    image as ximage,
    tool_result as xtool_result
)

# get_tool_call_type helps filter client-side vs server-side tools
try:
    from xai_sdk.tools import get_tool_call_type as x_get_tool_call_type
except Exception:
    x_get_tool_call_type = None


class Responses:
    def __init__(self, window=None):
        """
        Chat wrapper for xAI using Chat Responses via xai_sdk.

        :param window: Window instance
        """
        self.window = window
        self.input_tokens = 0
        # Image constraints
        self.allowed_mimes = {"image/jpeg", "image/png"}
        self.default_image_max_bytes = 10 * 1024 * 1024  # 10 MiB default

    # ---------- SEND (Chat Responses) ----------

    def send(self, context: BridgeContext, extra: Optional[Dict[str, Any]] = None):
        """
        Entry point for xAI chat/multimodal using Chat Responses (stateful).

        Streaming:
        - Uses xai_sdk chat.stream() (tuples of (response, chunk)).

        Non-stream:
        - Uses xai_sdk chat.sample().

        Client-side tools:
        - Prepared from app functions and mixed with server-side tools.
        - Only client-side tool calls are returned in ctx.tool_calls (server-side are executed on xAI).

        :param context: BridgeContext with all parameters
        :param extra: Extra parameters (not used)
        :return: SDK response object (non-stream) or an iterable for streaming (generator/iterator)
        """
        prompt = context.prompt
        system_prompt = context.system_prompt
        model_item = context.model
        functions = context.external_functions
        attachments = context.attachments
        multimodal_ctx = context.multimodal_ctx
        stream = context.stream
        history = context.history
        ctx = context.ctx or CtxItem()

        client = self.window.core.api.xai.get_client(context.mode, model_item)

        # Local input token estimate (best effort)
        self.reset_tokens()
        cnt_msgs = self._build_count_messages(prompt, system_prompt, model_item, history)
        self.input_tokens += self.window.core.tokens.from_messages(cnt_msgs, model_item.id)

        # Remote (server-side) tools and include flags
        rt_cfg = self.window.core.api.xai.remote.build_for_chat(
            model=model_item,
            stream=stream,
        )
        srv_tools = rt_cfg.get("tools", []) or []
        include = rt_cfg.get("include", []) or []
        use_encrypted = bool(rt_cfg.get("use_encrypted_content", False))
        max_turns = rt_cfg.get("max_turns", None)

        # Client-side tools (from app functions)
        client_tools = self.window.core.api.xai.tools.prepare_sdk_tools(functions)
        all_tools = (srv_tools + client_tools) if (srv_tools or client_tools) else None

        # Vision fallback if needed
        has_images = self._attachments_have_images(attachments)
        model_id = model_item.id
        if has_images and not self._is_vision_model(model_item):
            fb = self.window.core.config.get("xai_vision_fallback_model") or "grok-2-vision-latest"
            self.window.core.debug.info(f"[xai] Switching to vision model: {fb} (was: {model_id}) due to image input")
            model_id = fb

        # Store messages: false when images present (SDK guidance), otherwise configurable (default True)
        store_messages = True
        cfg_store = self.window.core.config.get("remote_tools.xai.store_messages")
        if cfg_store is not None:
            try:
                store_messages = bool(cfg_store)
            except Exception:
                pass
        if has_images:
            store_messages = False

        # previous_response_id from last history item or current ctx
        prev_id = self._detect_previous_response_id(history, ctx)

        # Create chat session in SDK
        chat_kwargs: Dict[str, Any] = {
            "model": model_id,
            "tools": all_tools,
            "include": (include if include else None),
            "store_messages": store_messages,
            "previous_response_id": prev_id,
        }
        if use_encrypted:
            chat_kwargs["use_encrypted_content"] = True
        if isinstance(max_turns, int) and max_turns > 0:
            chat_kwargs["max_turns"] = max_turns

        chat = client.chat.create(**chat_kwargs)

        # Append history (only when not continuing via previous_response_id)
        self.append_history_sdk(
            chat=chat,
            system_prompt=system_prompt,
            model=model_item,
            history=history if prev_id is None else None,  # do not duplicate when chaining
        )

        # If last turn contained client-side tool outputs, append them first
        self._append_tool_results_from_ctx(chat, history)

        # Append current user message (with images if any)
        self.append_current_user_sdk(
            chat=chat,
            prompt=prompt,
            attachments=attachments,
            multimodal_ctx=multimodal_ctx,
        )

        # STREAMING: return raw iterator (handler attaches ctx.msg_id)
        if stream:
            return chat.stream()

        # NON-STREAM
        response = chat.sample()
        return response

    # ---------- UNPACK (non-stream) ----------

    def unpack_response(self, mode: str, response, ctx: CtxItem):
        """
        Unpack non-streaming xAI response into ctx (text, tool calls, usage, citations, images).

        :param mode: mode (chat, etc)
        :param response: Response object from SDK or dict
        :param ctx: CtxItem to fill
        """
        # Output text
        out = ""
        try:
            out = getattr(response, "content", None) or ""
        except Exception:
            pass
        if not out and isinstance(response, dict):
            out = response.get("output_text") or ""

        ctx.output = (str(out or "")).strip()

        # Citations (list of urls)
        try:
            cits = getattr(response, "citations", None)
            if isinstance(cits, list) and cits:
                if ctx.urls is None:
                    ctx.urls = []
                for u in cits:
                    if isinstance(u, str) and (u.startswith("http://") or u.startswith("https://")):
                        if u not in ctx.urls:
                            ctx.urls.append(u)
        except Exception:
            pass

        # Tool calls: only return client-side tools to be executed by the app
        try:
            raw_calls = getattr(response, "tool_calls", None) or []
            client_side = []
            for tc in raw_calls:
                # filter client-side tool calls using SDK helper when available
                ttype = None
                if x_get_tool_call_type is not None:
                    try:
                        ttype = x_get_tool_call_type(tc)
                    except Exception:
                        ttype = None
                # treat as client-side when helper says so, or when object looks like function call
                if ttype == "client_side_tool" or self._looks_like_client_tool(tc):
                    fn = getattr(tc, "function", None)
                    name = getattr(fn, "name", "") if fn is not None else ""
                    args = getattr(fn, "arguments", "") if fn is not None else ""
                    if isinstance(args, (dict, list)):
                        try:
                            args = json.dumps(args, ensure_ascii=False)
                        except Exception:
                            args = str(args)
                    client_side.append({
                        "id": getattr(tc, "id", "") or "",
                        "type": "function",
                        "function": {"name": name or "", "arguments": args or ""},
                    })
            if client_side:
                ctx.tool_calls = client_side
        except Exception:
            pass

        # Usage
        try:
            usage = getattr(response, "usage", None)
            if isinstance(usage, dict):
                u = self._normalize_usage(usage)
            else:
                # Sometimes SDK exposes usage as object with attributes
                u = self._normalize_usage({
                    "prompt_tokens": getattr(usage, "prompt_tokens", 0) if usage else 0,
                    "completion_tokens": getattr(usage, "completion_tokens", 0) if usage else 0,
                    "reasoning_tokens": getattr(usage, "reasoning_tokens", 0) if usage else 0,
                })
            if u:
                ctx.set_tokens(u.get("in", 0), u.get("out", 0))
                if not isinstance(ctx.extra, dict):
                    ctx.extra = {}
                ctx.extra["usage"] = {
                    "vendor": "xai",
                    "input_tokens": u.get("in", 0),
                    "output_tokens": u.get("out", 0),
                    "reasoning_tokens": u.get("reasoning", 0),
                    "total_reported": u.get("total"),
                }
        except Exception:
            pass

        # Response ID
        try:
            rid = getattr(response, "id", None)
            if not rid and isinstance(response, dict):
                rid = response.get("id")
            if rid:
                ctx.msg_id = str(rid)
                if not isinstance(ctx.extra, dict):
                    ctx.extra = {}
                ctx.extra["xai_response_id"] = ctx.msg_id
        except Exception:
            pass

    # ---------- SDK message building helpers ----------

    def append_history_sdk(
        self,
        chat,
        system_prompt: Optional[str],
        model: ModelItem,
        history: Optional[List[CtxItem]] = None,
    ):
        """
        Append history to SDK chat (only when not using previous_response_id).

        :param chat: xai_sdk chat instance
        :param system_prompt: system prompt
        :param model: model item
        :param history: history items
        """
        if history is None:
            # initial turn, append system only (if any)
            if system_prompt:
                chat.append(xsystem(system_prompt))
            return

        # system prompt first (if any)
        if system_prompt:
            chat.append(xsystem(system_prompt))

        # history pairs
        if self.window.core.config.get('use_context'):
            used = self.window.core.tokens.from_user("", system_prompt or "")
            items = self.window.core.ctx.get_history(
                history, model.id, MODE_CHAT, used, self._fit_ctx(model),
            )
            for item in items:
                if item.final_input:
                    chat.append(xuser(str(item.final_input)))
                if item.final_output:
                    chat.append(xassistant(str(item.final_output)))

    def append_current_user_sdk(
        self,
        chat,
        prompt: str,
        attachments: Optional[Dict[str, AttachmentItem]],
        multimodal_ctx: Optional[MultimodalContext],
    ):
        """
        Append current user message with optional inline images.

        :param chat: xai_sdk chat instance
        :param prompt: user text
        :param attachments: attachments dict (images)
        :param multimodal_ctx: multimodal context (not used here)
        """
        parts = [str(prompt or "")]
        for img in self.window.core.api.xai.vision.build_images_for_chat(attachments):
            parts.append(ximage(img))
        chat.append(xuser(*parts))

    def _append_tool_results_from_ctx(self, chat, history: Optional[List[CtxItem]]):
        """
        Append tool results from the last ctx item when function-calling loop is active.
        """
        tool_call_native_enabled = self.window.core.config.get('func_call.native', False)
        if not (history and tool_call_native_enabled):
            return
        last = history[-1]
        if not (last.extra and isinstance(last.extra, dict)):
            return
        tool_output = last.extra.get("tool_output")
        if not (tool_output and isinstance(tool_output, list)):
            return

        for out in tool_output:
            # accept direct result value or whole dict
            if isinstance(out, dict) and "result" in out:
                chat.append(xtool_result(str(out["result"])))
            else:
                chat.append(xtool_result(str(out)))

    # ---------- legacy/local utils ----------

    def _fit_ctx(self, model: ModelItem) -> int:
        """
        Fit to max model tokens (uses model.ctx if present).
        """
        max_ctx_tokens = self.window.core.config.get('max_total_tokens')
        if model and model.ctx and 0 < model.ctx < max_ctx_tokens:
            max_ctx_tokens = model.ctx
        return max_ctx_tokens

    def _build_count_messages(
        self,
        prompt: str,
        system_prompt: str,
        model: ModelItem,
        history: Optional[List[CtxItem]] = None,
    ) -> List[dict]:
        """
        Build simple messages structure for local token estimation.
        """
        messages = []
        if system_prompt:
            messages.append({"role": "system", "content": system_prompt})

        if self.window.core.config.get('use_context'):
            used_tokens = self.window.core.tokens.from_user(prompt, system_prompt)
            items = self.window.core.ctx.get_history(
                history, model.id, MODE_CHAT, used_tokens, self._fit_ctx(model),
            )
            for item in items:
                if item.final_input:
                    messages.append({"role": "user", "content": str(item.final_input)})
                if item.final_output:
                    messages.append({"role": "assistant", "content": str(item.final_output)})

        messages.append({"role": "user", "content": str(prompt)})
        return messages

    def _normalize_usage(self, raw) -> Optional[dict]:
        """
        Normalize usage to a common dict: {'in','out','reasoning','total'}.
        Accepts either:
          - {'input_tokens','output_tokens','total_tokens'}
          - {'prompt_tokens','completion_tokens','total_tokens'}
        """
        if not isinstance(raw, dict):
            return None

        def _as_int(v) -> int:
            try:
                return int(v)
            except Exception:
                try:
                    return int(float(v))
                except Exception:
                    return 0

        in_tok = raw.get("input_tokens") if "input_tokens" in raw else raw.get("prompt_tokens")
        out_tok = raw.get("output_tokens") if "output_tokens" in raw else raw.get("completion_tokens")
        reasoning_tok = raw.get("reasoning_tokens", 0)
        tot = raw.get("total_tokens")

        i = _as_int(in_tok or 0)
        o = _as_int(out_tok or 0)
        r = _as_int(reasoning_tok or 0)
        t = _as_int(tot if tot is not None else (i + o + r))
        return {"in": i, "out": max(0, t - i) if t else o, "reasoning": r, "total": t}

    def _attachments_have_images(self, attachments: Optional[Dict[str, AttachmentItem]]) -> bool:
        """
        Detect if attachments contain at least one image file.
        """
        if not attachments:
            return False
        for _, att in attachments.items():
            try:
                if att.path and self.window.core.api.xai.vision.is_image(att.path):
                    return True
            except Exception:
                continue
        return False

    def _is_vision_model(self, model: ModelItem) -> bool:
        """
        Heuristic check for vision-capable model IDs.
        """
        model_id = (model.id if model and model.id else "").strip()
        if not model or not model_id:
            return False
        if model.is_image_input():
            return True
        mid = model_id.lower()
        return ("vision" in mid) or ("-v" in mid and "grok" in mid)

    def _looks_like_client_tool(self, tc_obj) -> bool:
        """
        Best-effort detection of a client-side tool call when helper is not available.
        """
        try:
            fn = getattr(tc_obj, "function", None)
            if fn is None:
                return False
            name = getattr(fn, "name", None)
            # arguments can be str/dict/list; presence indicates structured call
            has_args = hasattr(fn, "arguments")
            return isinstance(name, str) and name != "" and has_args
        except Exception:
            return False

    def _detect_previous_response_id(self, history: Optional[List[CtxItem]], ctx: CtxItem) -> Optional[str]:
        """
        Return last response id from history or current ctx when available.
        """
        try:
            if history and len(history) > 0:
                last = history[-1]
                if last and last.msg_id:
                    return str(last.msg_id)
        except Exception:
            pass
        try:
            if ctx and ctx.msg_id:
                return str(ctx.msg_id)
        except Exception:
            pass
        return None

    def reset_tokens(self):
        """Reset input tokens counter."""
        self.input_tokens = 0

    def get_used_tokens(self) -> int:
        """
        Return the locally estimated input tokens count.
        """
        return self.input_tokens

    # ---------- helpers for quick_call compatibility ----------

    def quick_collect_response_id(self, response, ctx: CtxItem):
        """
        Set response id and xai_response_id after non-stream sample() in quick paths.

        :param response: Response object from SDK or dict
        :param ctx: CtxItem to fill
        """
        try:
            rid = getattr(response, "id", None)
            if rid:
                ctx.msg_id = str(rid)
                if not isinstance(ctx.extra, dict):
                    ctx.extra = {}
                ctx.extra["xai_response_id"] = ctx.msg_id
        except Exception:
            pass


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.17 19:00:00                  #
# ================================================== #

from typing import Optional, Dict, List

from llama_index.core.base.embeddings.base import BaseEmbedding
from llama_index.core.llms.llm import BaseLLM as LlamaBaseLLM

from pygpt_net.core.types import (
    MODE_LLAMA_INDEX,
)
from pygpt_net.provider.llms.base import BaseLLM
from pygpt_net.item.model import ModelItem


class OpenRouterLLM(BaseLLM):
    def __init__(self, *args, **kwargs):
        super(OpenRouterLLM, self).__init__(*args, **kwargs)
        self.id = "open_router"
        self.name = "OpenRouter"
        self.type = [MODE_LLAMA_INDEX, "embeddings"]

    def get_embeddings_model(
            self,
            window,
            config: Optional[List[Dict]] = None
    ) -> BaseEmbedding:
        """
        Return provider instance for embeddings

        :param window: window instance
        :param config: config keyword arguments list
        :return: Embedding provider instance
        """
        from llama_index.embeddings.openai_like import OpenAILikeEmbedding
        args = {}
        if config is not None:
            args = self.parse_args({
                "args": config,
            }, window)
        if "api_key" not in args:
            args["api_key"] = window.core.config.get("api_key_open_router", "")
        if "api_base" not in args:
            args["api_base"] = window.core.config.get("api_endpoint_open_router", "")
        if "model" in args and "model_name" not in args:
            args["model_name"] = args.pop("model")
        args = self.inject_llamaindex_http_clients(args, window.core.config)
        return OpenAILikeEmbedding(**args)

    def llama(
            self,
            window,
            model: ModelItem,
            stream: bool = False
    ) -> LlamaBaseLLM:
        """
        Return LLM provider instance for llama

        :param window: window instance
        :param model: model instance
        :param stream: stream mode
        :return: LLM provider instance
        """
        from llama_index.llms.openai_like import OpenAILike
        args = self.parse_args(model.llama_index, window)
        if "model" not in args:
            args["model"] = model.id
        if "api_key" not in args:
            args["api_key"] = window.core.config.get("api_key_open_router", "")
        if "api_base" not in args:
            args["api_base"] = window.core.config.get("api_endpoint_open_router", "")
        if "is_chat_model" not in args:
            args["is_chat_model"] = True
        if "is_function_calling_model" not in args:
            args["is_function_calling_model"] = model.tool_calls
        args = self.inject_llamaindex_http_clients(args, window.core.config)
        if model:
            args["model"] = window.core.models.get_openrouter_model(model)
        return OpenAILike(**args)

    def get_models(
            self,
            window,
    ) -> List[Dict]:
        """
        Return list of models for the provider

        :param window: window instance
        :return: list of models
        """
        items = []
        client = self.get_client(window)
        models_list = client.models.list()
        if models_list.data:
            for item in models_list.data:
                items.append({
                    "id": item.id,
                    "name": item.id,
                })
        return items


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.24 23:00:00                  #
# ================================================== #

from PySide6.QtGui import QAction, QIcon
from PySide6.QtWidgets import QMenu

from pygpt_net.utils import trans


class Theme:
    def __init__(self, window=None):
        """
        Menu setup

        :param window: Window instance
        """
        self.window = window
        self._loaded = False

    def _on_toggle_tooltips(self, checked=False):
        self.window.controller.theme.toggle_option('layout.tooltips')

    def _open_settings(self, checked=False):
        self.window.controller.settings.open_section('layout')

    def setup(self):
        """Setup theme menu"""
        w = self.window
        m = w.ui.menu

        if self._loaded:
            act = m.get('theme.tooltips')
            if isinstance(act, QAction):
                act.setChecked(w.core.config.get('layout.tooltips'))
            return

        m['theme'] = {}
        m['theme_style'] = {}
        m['theme_syntax'] = {}
        m['theme.layout.density'] = {}
        m['menu.theme'] = QMenu(trans("menu.theme"), w)

        m['theme.style'] = QMenu(trans("menu.theme.style"), w)

        m['theme.dark'] = QMenu(trans("menu.theme.dark"), w)
        m['theme.light'] = QMenu(trans("menu.theme.light"), w)
        m['theme.syntax'] = QMenu(trans("menu.theme.syntax"), w)

        m['theme.density'] = QMenu(trans("menu.theme.density"), w)

        m['theme.tooltips'] = QAction(trans("menu.theme.tooltips"), w, checkable=True)
        m['theme.tooltips'].triggered.connect(self._on_toggle_tooltips)
        m['theme.tooltips'].setChecked(w.core.config.get('layout.tooltips'))

        m['theme.settings'] = QAction(QIcon(":/icons/settings_filled.svg"),
                                      trans("menu.theme.settings"), w)
        m['theme.settings'].setMenuRole(QAction.MenuRole.NoRole)
        m['theme.settings'].triggered.connect(self._open_settings)

        menu_theme = m['menu.theme']
        menu_theme.addMenu(m['theme.style'])
        menu_theme.addMenu(m['theme.dark'])
        menu_theme.addMenu(m['theme.light'])
        menu_theme.addMenu(m['theme.syntax'])
        menu_theme.addMenu(m['theme.density'])
        menu_theme.addAction(m['theme.tooltips'])
        menu_theme.addAction(m['theme.settings'])

        self._loaded = True


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.12.28 04:00:00                  #
# ================================================== #

from PySide6.QtCore import Qt, QItemSelectionModel
from PySide6.QtGui import QAction, QIcon
from PySide6.QtWidgets import QMenu, QAbstractItemView

from pygpt_net.ui.widget.lists.base import BaseList
from pygpt_net.utils import trans


class ProfileList(BaseList):
    def __init__(self, window=None, id=None):
        """
        Profile menu (in editor)

        :param window: main window
        :param id: parent id
        """
        super(ProfileList, self).__init__(window)
        self.window = window
        self.id = id

        # Enable row-based multi-select with native Ctrl/Shift gestures
        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.setSelectionMode(QAbstractItemView.ExtendedSelection)

        # Context menu restore helper
        self._backup_selection = None
        self.restore_after_ctx_menu = True

    def _selected_rows(self) -> list[int]:
        """Return list of selected row numbers."""
        try:
            return [ix.row() for ix in self.selectionModel().selectedRows()]
        except Exception:
            return []

    def _has_multi_selection(self) -> bool:
        """Return True when multiple rows are selected."""
        try:
            return len(self.selectionModel().selectedRows()) > 1
        except Exception:
            return False

    def mousePressEvent(self, event):
        """
        Mouse press event
        - Ctrl: let Qt toggle selection (virtual; no business action).
        - Shift: let Qt range-select (virtual).
        - If multiple are selected: a single left click anywhere clears selection.
        """
        if event.button() == Qt.LeftButton and (event.modifiers() & Qt.ControlModifier):
            idx = self.indexAt(event.pos())
            if idx.isValid():
                super(ProfileList, self).mousePressEvent(event)  # native toggle
            else:
                event.accept()
            return

        if event.button() == Qt.LeftButton and (event.modifiers() & Qt.ShiftModifier):
            idx = self.indexAt(event.pos())
            if idx.isValid():
                super(ProfileList, self).mousePressEvent(event)  # native range
            else:
                event.accept()
            return

        if event.button() == Qt.LeftButton:
            index = self.indexAt(event.pos())
            # Clear multi-selection with a single click (also on empty area)
            if self._has_multi_selection():
                sel_model = self.selectionModel()
                sel_model.clearSelection()
                if not index.isValid():
                    event.accept()
                    return
            # Default selection behavior
            super(ProfileList, self).mousePressEvent(event)
            return

        super(ProfileList, self).mousePressEvent(event)

    def click(self, val):
        pass

    def contextMenuEvent(self, event):
        """
        Context menu event
        - Shows menu on a row, or on empty area when multi-selection is active.
        - For multi-selection, actions pass list[int] of selected rows.
        - For single selection, actions pass single int (legacy behavior).
        """
        actions = {}
        actions['use'] = QAction(QIcon(":/icons/check.svg"), trans('action.use'), self)
        actions['edit'] = QAction(QIcon(":/icons/edit.svg"), trans('action.edit'), self)
        actions['duplicate'] = QAction(QIcon(":/icons/copy.svg"), trans('action.duplicate'), self)
        actions['reset'] = QAction(QIcon(":/icons/close.svg"), trans('action.reset'), self)
        actions['delete'] = QAction(QIcon(":/icons/delete.svg"), trans('action.profile.delete'), self)
        actions['delete_all'] = QAction(QIcon(":/icons/delete.svg"), trans('action.profile.delete_all'), self)

        # Hit test
        index = self.indexAt(event.pos())
        idx = index.row() if index.isValid() else -1

        # Current selection state
        sel_model = self.selectionModel()
        selected_indexes = list(sel_model.selectedRows()) if sel_model else []
        selected_rows = [ix.row() for ix in selected_indexes]
        multi = len(selected_rows) > 1

        # Allow menu on empty area only when multi-selection is active
        if not index.isValid() and not multi:
            return

        # If right-click outside the current multi-selection, temporarily select clicked row
        backup_selection = None
        if index.isValid():
            if multi and idx in selected_rows:
                backup_selection = None  # keep selection as-is
            else:
                backup_selection = list(sel_model.selectedIndexes())
                sel_model.clearSelection()
                sel_model.select(index, QItemSelectionModel.Select | QItemSelectionModel.Rows)
                selected_rows = [idx]
                multi = False

        menu = QMenu(self)

        if not multi:
            # Single: keep legacy behavior (pass single index via event)
            menu.addAction(actions['edit'])
            actions['edit'].triggered.connect(lambda: self.action_edit(event))

            menu.addAction(actions['use'])
            actions['use'].triggered.connect(lambda: self.action_use(event))

            menu.addAction(actions['duplicate'])
            actions['duplicate'].triggered.connect(lambda: self.action_duplicate(event))

            menu.addAction(actions['reset'])
            actions['reset'].triggered.connect(lambda: self.action_reset(event))

            menu.addAction(actions['delete'])
            actions['delete'].triggered.connect(lambda: self.action_delete(event))

            menu.addAction(actions['delete_all'])
            actions['delete_all'].triggered.connect(lambda: self.action_delete_all(event))
        else:
            # Multi: keep only actions that make sense in bulk; Use/Edit/Duplicate are single-only
            rows = list(selected_rows)

            menu.addAction(actions['reset'])
            actions['reset'].triggered.connect(lambda checked=False, r=rows: self._action_reset_multi(r))

            menu.addAction(actions['delete'])
            actions['delete'].triggered.connect(lambda checked=False, r=rows: self._action_delete_multi(r))

            menu.addAction(actions['delete_all'])
            actions['delete_all'].triggered.connect(lambda checked=False, r=rows: self._action_delete_all_multi(r))

        # Show menu
        if index.isValid() or multi:
            menu.exec_(event.globalPos())

        # Restore selection after context menu if it was temporarily changed
        if backup_selection is not None and self.restore_after_ctx_menu:
            sel_model.clearSelection()
            for i in backup_selection:
                sel_model.select(i, QItemSelectionModel.Select | QItemSelectionModel.Rows)
        self._backup_selection = None
        self.restore_after_ctx_menu = True

    # ---------- Single-item handlers (legacy; keep unchanged signatures) ----------

    def action_use(self, event):
        """
        Use action handler

        :param event: mouse event
        """
        item = self.indexAt(event.pos())
        idx = item.row()
        if idx >= 0:
            self.window.controller.settings.profile.select_by_idx(idx)

    def action_duplicate(self, event):
        """
        Duplicate action handler

        :param event: mouse event
        """
        item = self.indexAt(event.pos())
        idx = item.row()
        if idx >= 0:
            self.window.controller.settings.profile.duplicate_by_idx(idx)

    def action_reset(self, event):
        """
        Reset action handler

        :param event: mouse event
        """
        item = self.indexAt(event.pos())
        idx = item.row()
        if idx >= 0:
            self.window.controller.settings.profile.reset_by_idx(idx)

    def action_delete(self, event):
        """
        Delete action handler

        :param event: mouse event
        """
        item = self.indexAt(event.pos())
        idx = item.row()
        if idx >= 0:
            self.window.controller.settings.profile.delete_by_idx(idx)

    def action_delete_all(self, event):
        """
        Delete all action handler

        :param event: mouse event
        """
        item = self.indexAt(event.pos())
        idx = item.row()
        if idx >= 0:
            self.window.controller.settings.profile.delete_all_by_idx(idx)

    def action_edit(self, event):
        """
        Edit action handler

        :param event: mouse event
        """
        item = self.indexAt(event.pos())
        idx = item.row()
        if idx >= 0:
            self.window.controller.settings.profile.edit_by_idx(idx)

    # ---------- Multi-item handlers (new; pass list[int]) ----------

    def _action_reset_multi(self, rows: list[int]):
        if rows:
            self.window.controller.settings.profile.reset_by_idx(list(rows))

    def _action_delete_multi(self, rows: list[int]):
        if rows:
            self.window.controller.settings.profile.delete_by_idx(list(rows))

    def _action_delete_all_multi(self, rows: list[int]):
        if rows:
            self.window.controller.settings.profile.delete_all_by_idx(list(rows))


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.07.16 02:00:00                  #
# ================================================== #

import os
import re

from pygpt_net.item.ctx import CtxItem


class Parser:
    def __init__(self, window=None):
        """
        Filesystem parser

        :param window: Window instance
        """
        self.window = window

    def extract_data_paths(self, text: str) -> list:
        """
        Extract file paths from text that contain 'data' segment (case-insensitive).
        Supports quoted and unquoted paths, POSIX/Windows, and ignores URLs.
        """
        if not text:
            return []

        def is_data_path(p: str) -> bool:
            # 'data' (case-insensitive)
            return re.search(r"(?i)(?:^|[\\/])data(?:[\\/]|$)", p) is not None

        def is_url(p: str) -> bool:
            return re.match(r"^[a-z][a-z0-9+.-]*://", p, re.I) is not None

        results = []

        quoted_pat = re.compile(r"(?P<q>['\"])(?P<p>(?:[A-Za-z]:)?[\\/](?:(?!\1).)+?)\1")
        for m in quoted_pat.finditer(text):
            p = m.group("p").strip()
            if not is_url(p) and is_data_path(p):
                results.append(p)

        unquoted_pat = re.compile(r"(?P<p>(?:[A-Za-z]:)?(?:[\\/][^\s'\"),;]+)+)")
        for m in unquoted_pat.finditer(text):
            p = m.group("p").strip()
            if not is_url(p) and is_data_path(p):
                results.append(p)

        seen = set()
        out = []
        for p in results:
            if p not in seen:
                seen.add(p)
                out.append(p)
        return out

    def extract_data_files(self, ctx: "CtxItem", response: str) -> list:
        """
        Extract files from tool outputs and return list of file paths.
        """
        if response is None:
            return []
        images_list = []
        local_data_dir = self.window.core.config.get_user_dir('data')
        raw_paths = self.extract_data_paths(response)

        def replace_with_local(path):
            """
            Replace the path with local data directory path.
            """
            segments = re.split(r"[\\/]+", path)
            # case-insensitive find of 'data'
            data_index = next((i for i, s in enumerate(segments) if s.lower() == "data"), None)
            if data_index is None:
                return path
            tail = segments[data_index + 1:]
            new_path = os.path.join(local_data_dir, *tail) if tail else local_data_dir
            return new_path

        processed_paths = []
        for file in raw_paths:
            new_file = replace_with_local(file)
            if new_file not in processed_paths:
                processed_paths.append(new_file)

        for path in processed_paths:
            ext = os.path.splitext(path)[1].lower().lstrip(".")
            if ext in ["png", "jpg", "jpeg", "gif", "bmp", "tiff", "webp"]:
                if path not in images_list:
                    images_list.append(path)

        local_images = self.window.core.filesystem.make_local_list(images_list)
        ctx.files = processed_paths
        ctx.images = local_images
        return processed_paths



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.12.25 20:00:00                  #
# ================================================== #

from typing import Any, Dict, List


class CheckboxList:
    def __init__(self, window=None):
        """
        CheckboxList field handler

        :param window: Window instance
        """
        self.window = window

    def apply(
            self,
            parent_id: str,
            key: str,
            option: Dict[str, Any]
    ):
        """
        Apply values to checkboxes

        :param parent_id: Options parent ID
        :param key: Option key
        :param option: Option data
        """
        if "value" not in option:
            return
        value = option["value"]
        selection = {s.strip() for s in value.split(",")} if isinstance(value, str) else set()
        selection.discard("")

        ui = self.window.ui
        cfg_parent = ui.config.get(parent_id)
        if not cfg_parent:
            return
        entry = cfg_parent.get(key)
        if entry is None or not hasattr(entry, "boxes"):
            return
        boxes = entry.boxes

        for name, cb in boxes.items():
            if cb is None:
                continue
            desired = name in selection
            if cb.isChecked() != desired:
                cb.setChecked(desired)

    def on_update(
            self,
            parent_id: str,
            key: str,
            option: dict,
            value: Any,
            subkey: str = None,
            hooks: bool = True
    ):
        """
        Event: on update checkbox value

        :param parent_id: Options parent ID
        :param key: Option key
        :param option: Option data
        :param value: Option value
        :param subkey: Subkey for specific checkbox
        :param hooks: Run hooks
        """
        if hooks:
            ui = self.window.ui
            hook_name = f"update.{parent_id}.{key}"
            if ui.has_hook(hook_name):
                hook = ui.get_hook(hook_name)
                if hook:
                    try:
                        hook(key, value, 'bool_list')
                    except Exception as e:
                        self.window.core.debug.log(e)


    def on_select_all(
            self,
            parent_id: str,
            key: str,
            option: dict
    ):
        """
        Event: select all checkboxes

        :param parent_id: Options parent ID
        :param key: Option key
        :param option: Option data
        """
        ui = self.window.ui
        cfg_parent = ui.config.get(parent_id)
        if not cfg_parent:
            return
        entry = cfg_parent.get(key)
        if entry is None or not hasattr(entry, "boxes"):
            return
        boxes = entry.boxes

        mode = "unselect_all"

        for name, cb in boxes.items():
            if cb is not None and not cb.isChecked():
                mode = "select_all"

        for name, cb in boxes.items():
            if cb is not None:
                cb.setChecked(mode == "select_all")

    def get_value(
            self,
            parent_id: str,
            key: str,
            option: Dict[str, Any]
    ) -> str:
        """
        Get checkbox values as comma-separated string

        :param parent_id: Parent ID
        :param key: Option key
        :param option: Option data dict
        :return: Option value
        """
        ui = self.window.ui
        cfg_parent = ui.config.get(parent_id)
        if not cfg_parent:
            return ""
        entry = cfg_parent.get(key)
        if entry is None or not hasattr(entry, "boxes"):
            return ""
        boxes = entry.boxes
        return ",".join(name for name, cb in boxes.items() if cb is not None and cb.isChecked())

    def update_list(
            self,
            parent_id: str,
            key: str,
            items: List[Dict]
    ):
        """
        Update combobox items

        :param parent_id: Options parent ID
        :param key: Option key
        :param items: Items dict
        """
        self.window.ui.config[parent_id][key].update_boxes_list(items)